from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional, Union

import pytest

from huggingface_hub.utils import DEFAULT_IGNORE_PATTERNS, filter_repo_objects


@dataclass
class DummyObject:
    path: Path


DUMMY_FILES = ["not_hidden.pdf", "profile.jpg", ".hidden.pdf", ".hidden_picture.png"]
DUMMY_PATHS = [Path(path) for path in DUMMY_FILES]
DUMMY_OBJECTS = [DummyObject(path=path) for path in DUMMY_FILES]


class TestPathsUtils:
    def test_get_all_pdfs(self) -> None:
        """Get all PDFs even hidden ones."""
        self._check(
            items=DUMMY_FILES,
            expected_items=["not_hidden.pdf", ".hidden.pdf"],
            allow_patterns=["*.pdf"],
        )

    def test_get_all_pdfs_except_hidden(self) -> None:
        """Get all PDFs except hidden ones."""
        self._check(
            items=DUMMY_FILES,
            expected_items=["not_hidden.pdf"],
            allow_patterns=["*.pdf"],
            ignore_patterns=[".*"],
        )

    def test_get_all_pdfs_except_hidden_using_single_pattern(self) -> None:
        """Get all PDFs except hidden ones, using single pattern."""
        self._check(
            items=DUMMY_FILES,
            expected_items=["not_hidden.pdf"],
            allow_patterns="*.pdf",  # not a list
            ignore_patterns=".*",  # not a list
        )

    def test_get_all_images(self) -> None:
        """Get all images."""
        self._check(
            items=DUMMY_FILES,
            expected_items=["profile.jpg", ".hidden_picture.png"],
            allow_patterns=["*.png", "*.jpg"],
        )

    def test_get_all_images_except_hidden_from_paths(self) -> None:
        """Get all images except hidden ones, from Path list."""
        self._check(
            items=DUMMY_PATHS,
            expected_items=[Path("profile.jpg")],
            allow_patterns=["*.png", "*.jpg"],
            ignore_patterns=".*",
        )

    def test_get_all_images_except_hidden_from_objects(self) -> None:
        """Get all images except hidden ones, from object list."""
        self._check(
            items=DUMMY_OBJECTS,
            expected_items=[DummyObject(path="profile.jpg")],
            allow_patterns=["*.png", "*.jpg"],
            ignore_patterns=".*",
            key=lambda x: x.path,
        )

    def test_filter_objects_key_not_provided(self) -> None:
        """Test ValueError is raised if filtering non-string objects."""
        with pytest.raises(ValueError, match="Please provide `key` argument"):
            list(
                filter_repo_objects(
                    items=DUMMY_OBJECTS,
                    allow_patterns=["*.png", "*.jpg"],
                    ignore_patterns=".*",
                )
            )

    def test_filter_object_with_folder(self) -> None:
        self._check(
            items=[
                "file.txt",
                "lfs.bin",
                "path/to/file.txt",
                "path/to/lfs.bin",
                "nested/path/to/file.txt",
                "nested/path/to/lfs.bin",
            ],
            expected_items=["path/to/file.txt", "path/to/lfs.bin"],
            allow_patterns=["path/to/"],
        )

    def test_filter_object_with_empty_pattern(self) -> None:
        # An empty pattern must not raise: `_add_wildcard_to_directories("")` indexed `pattern[-1]`.
        self._check(
            items=["file.txt", "lfs.bin"],
            expected_items=[],
            allow_patterns=[""],
        )
        self._check(
            items=["file.txt", "lfs.bin"],
            expected_items=["file.txt", "lfs.bin"],
            ignore_patterns=[""],
        )

    def test_filter_is_case_sensitive(self) -> None:
        """Pattern matching is case-sensitive.
        Regression test for https://github.com/huggingface/huggingface_hub/issues/4434.
        """
        # Uppercase pattern matches only the uppercase path, not the lowercase one.
        self._check(
            items=["README.md", "notes.MD"],
            expected_items=["notes.MD"],
            allow_patterns=["*.MD"],
        )
        # Lowercase pattern matches only the lowercase path, not the uppercase one.
        self._check(
            items=["README.md", "notes.MD"],
            expected_items=["README.md"],
            allow_patterns=["*.md"],
        )
        # Case-sensitivity also applies to the ignore list.
        self._check(
            items=["keep.txt", "drop.LOG", "keep.log"],
            expected_items=["keep.txt", "keep.log"],
            ignore_patterns=["*.LOG"],
        )

    def test_backslashes_are_treated_as_path_separators(self) -> None:
        """Windows-style patterns (e.g. built with `os.path.join`) match `/`-separated repo paths.

        Regression test for https://github.com/huggingface/huggingface_hub/pull/4506.
        """
        # Backslash-separated pattern matches forward-slash paths.
        self._check(
            items=["unet/config.json", "vae/config.json", "unet/model.bin"],
            expected_items=["unet/config.json"],
            allow_patterns=["unet\\config.json"],
        )
        # Backslash-separated paths (e.g. local Windows paths) match forward-slash patterns.
        self._check(
            items=["unet\\config.json", "unet\\model.bin"],
            expected_items=["unet\\config.json"],
            allow_patterns=["unet/*.json"],
        )
        # Also applies to the ignore list.
        self._check(
            items=["unet/config.json", "unet/model.bin"],
            expected_items=["unet/model.bin"],
            ignore_patterns=["unet\\*.json"],
        )

    def test_key_returning_path_object(self) -> None:
        """A custom `key` returning a `Path` (not `str`) must work."""
        self._check(
            items=[DummyObject(path=Path("unet/config.json")), DummyObject(path=Path("unet/model.bin"))],
            expected_items=[DummyObject(path=Path("unet/config.json"))],
            allow_patterns=["unet/*.json"],
            key=lambda x: x.path,
        )

    def _check(
        self,
        items: list[Any],
        expected_items: list[Any],
        allow_patterns: Optional[Union[list[str], str]] = None,
        ignore_patterns: Optional[Union[list[str], str]] = None,
        key: Optional[Callable[[Any], str]] = None,
    ) -> None:
        """Run `filter_repo_objects` and check output against expected result."""
        assert (
            list(
                filter_repo_objects(
                    items=items,
                    allow_patterns=allow_patterns,
                    ignore_patterns=ignore_patterns,
                    key=key,
                )
            )
            == expected_items
        )


class TestDefaultIgnorePatterns:
    PATHS_TO_IGNORE = [
        ".git",
        ".git/file.txt",
        ".git/folder/file.txt",
        "path/to/folder/.git",
        "path/to/folder/.git/file.txt",
        "path/to/.git/folder/file.txt",
        ".cache/huggingface",
        ".cache/huggingface/file.txt",
        ".cache/huggingface/folder/file.txt",
        "path/to/.cache/huggingface",
        "path/to/.cache/huggingface/file.txt",
    ]

    VALID_PATHS = [
        ".gitignore",
        "path/foo.git/file.txt",
        "path/.git_bar/file.txt",
        "path/to/file.git",
        "file.huggingface",
        "path/file.huggingface",
        ".cache/huggingface_folder",
        ".cache/huggingface_folder/file.txt",
    ]

    def test_exclude_git_folder(self):
        filtered_paths = filter_repo_objects(
            items=self.PATHS_TO_IGNORE + self.VALID_PATHS, ignore_patterns=DEFAULT_IGNORE_PATTERNS
        )
        assert list(filtered_paths) == self.VALID_PATHS



# Copyright 2026-present, the HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the centralized HF URI parser ('huggingface_hub.utils._hf_uris')."""

import pytest

from huggingface_hub import constants
from huggingface_hub.errors import HfUriError
from huggingface_hub.utils import HfMount, HfUri, parse_hf_mount, parse_hf_uri


# ---------------------------------------------------------------------------
# HfUri success cases: (uri, expected_HfUri, expected_roundtrip)
# ---------------------------------------------------------------------------
URI_SUCCESS_CASES: list[tuple[str, HfUri, str]] = [
    # --- Models ----------------------------------------------------------------
    # Namespaced model (implicit type prefix)
    (
        "hf://my-org/my-model",
        HfUri(type="model", id="my-org/my-model"),
        "hf://models/my-org/my-model",
    ),
    # Namespaced model (explicit type prefix)
    (
        "hf://models/my-org/my-model",
        HfUri(type="model", id="my-org/my-model"),
        "hf://models/my-org/my-model",
    ),
    # Model with path inside the repo
    (
        "hf://models/my-org/my-model/config.json",
        HfUri(type="model", id="my-org/my-model", path_in_repo="config.json"),
        "hf://models/my-org/my-model/config.json",
    ),
    # Model with explicit revision
    (
        "hf://my-org/my-model@v1.0",
        HfUri(type="model", id="my-org/my-model", revision="v1.0"),
        "hf://models/my-org/my-model@v1.0",
    ),
    # Model with revision and path
    (
        "hf://models/my-org/my-model@dev/sub/file.bin",
        HfUri(type="model", id="my-org/my-model", revision="dev", path_in_repo="sub/file.bin"),
        "hf://models/my-org/my-model@dev/sub/file.bin",
    ),
    # --- Datasets --------------------------------------------------------------
    (
        "hf://datasets/my-org/my-dataset",
        HfUri(type="dataset", id="my-org/my-dataset"),
        "hf://datasets/my-org/my-dataset",
    ),
    # Dataset with revision and path
    (
        "hf://datasets/my-org/my-dataset@v1/train.csv",
        HfUri(type="dataset", id="my-org/my-dataset", revision="v1", path_in_repo="train.csv"),
        "hf://datasets/my-org/my-dataset@v1/train.csv",
    ),
    # --- Spaces ----------------------------------------------------------------
    (
        "hf://spaces/user/my-space",
        HfUri(type="space", id="user/my-space"),
        "hf://spaces/user/my-space",
    ),
    (
        "hf://spaces/user/my-space@main",
        HfUri(type="space", id="user/my-space", revision="main"),
        "hf://spaces/user/my-space@main",
    ),
    # --- Kernels ---------------------------------------------------------------
    (
        "hf://kernels/my-org/my-kernel",
        HfUri(type="kernel", id="my-org/my-kernel"),
        "hf://kernels/my-org/my-kernel",
    ),
    # --- Buckets ---------------------------------------------------------------
    (
        "hf://buckets/my-org/my-bucket",
        HfUri(type="bucket", id="my-org/my-bucket"),
        "hf://buckets/my-org/my-bucket",
    ),
    (
        "hf://buckets/my-org/my-bucket/sub/path",
        HfUri(type="bucket", id="my-org/my-bucket", path_in_repo="sub/path"),
        "hf://buckets/my-org/my-bucket/sub/path",
    ),
    # Trailing slashes are tolerated and stripped.
    (
        "hf://buckets/my-org/my-bucket/",
        HfUri(type="bucket", id="my-org/my-bucket"),
        "hf://buckets/my-org/my-bucket",
    ),
    # --- Special revisions (refs/pr/N, refs/convert/parquet) -------------------
    # Special PR revision: must be matched eagerly even though it contains '/'
    (
        "hf://datasets/foo/bar@refs/pr/10/file.csv",
        HfUri(type="dataset", id="foo/bar", revision="refs/pr/10", path_in_repo="file.csv"),
        "hf://datasets/foo/bar@refs/pr/10/file.csv",
    ),
    # Special convert revision
    (
        "hf://datasets/foo/bar@refs/convert/parquet/data.parquet",
        HfUri(
            type="dataset",
            id="foo/bar",
            revision="refs/convert/parquet",
            path_in_repo="data.parquet",
        ),
        "hf://datasets/foo/bar@refs/convert/parquet/data.parquet",
    ),
    # Convert ref name with hyphen (and dot) — must not be split at the hyphen.
    (
        "hf://datasets/foo/bar@refs/convert/parquet-v2/data.parquet",
        HfUri(
            type="dataset",
            id="foo/bar",
            revision="refs/convert/parquet-v2",
            path_in_repo="data.parquet",
        ),
        "hf://datasets/foo/bar@refs/convert/parquet-v2/data.parquet",
    ),
    (
        "hf://datasets/foo/bar@refs/convert/duckdb.v1/data.db",
        HfUri(
            type="dataset",
            id="foo/bar",
            revision="refs/convert/duckdb.v1",
            path_in_repo="data.db",
        ),
        "hf://datasets/foo/bar@refs/convert/duckdb.v1/data.db",
    ),
    # URL-encoded special revision
    (
        "hf://datasets/foo/bar@refs%2Fpr%2F10/file.csv",
        HfUri(type="dataset", id="foo/bar", revision="refs/pr/10", path_in_repo="file.csv"),
        "hf://datasets/foo/bar@refs/pr/10/file.csv",
    ),
    # Branch name containing '/' (e.g. 'feature/foo'). Must be URL-encoded so that the
    # parser does not split it at the first '/' and treat the rest as a path-in-repo.
    (
        "hf://my-org/my-model@feature%2Ffoo/config.json",
        HfUri(type="model", id="my-org/my-model", revision="feature/foo", path_in_repo="config.json"),
        "hf://models/my-org/my-model@feature%2Ffoo/config.json",
    ),
    # Special revision with no path after it
    (
        "hf://my-org/my-model@refs/pr/3",
        HfUri(type="model", id="my-org/my-model", revision="refs/pr/3"),
        "hf://models/my-org/my-model@refs/pr/3",
    ),
    # --- '@' in filenames (literal, not a revision separator) -----------------
    # Bucket with '@' in filename
    (
        "hf://buckets/my-org/my-bucket/file@1.txt",
        HfUri(type="bucket", id="my-org/my-bucket", path_in_repo="file@1.txt"),
        "hf://buckets/my-org/my-bucket/file@1.txt",
    ),
    # Bucket with '@' in nested path
    (
        "hf://buckets/my-org/my-bucket/sub/data@v2.bin",
        HfUri(type="bucket", id="my-org/my-bucket", path_in_repo="sub/data@v2.bin"),
        "hf://buckets/my-org/my-bucket/sub/data@v2.bin",
    ),
    # Repo with '@' in filename (no revision)
    (
        "hf://models/my-org/my-model/file@1.txt",
        HfUri(type="model", id="my-org/my-model", path_in_repo="file@1.txt"),
        "hf://models/my-org/my-model/file@1.txt",
    ),
    # Repo with '@' in nested path (no revision, implicit model type)
    (
        "hf://my-org/my-model/sub/data@v2.bin",
        HfUri(type="model", id="my-org/my-model", path_in_repo="sub/data@v2.bin"),
        "hf://models/my-org/my-model/sub/data@v2.bin",
    ),
    # Repo with explicit revision AND '@' in filename
    (
        "hf://models/my-org/my-model@main/file@1.txt",
        HfUri(type="model", id="my-org/my-model", revision="main", path_in_repo="file@1.txt"),
        "hf://models/my-org/my-model@main/file@1.txt",
    ),
    # '@' right after namespace/name is still a revision separator
    (
        "hf://a/b/c@v1",
        HfUri(type="model", id="a/b", path_in_repo="c@v1"),
        "hf://models/a/b/c@v1",
    ),
]


# ---------------------------------------------------------------------------
# HfUri failure cases: (uri, error_substring)
# ---------------------------------------------------------------------------
URI_FAILURE_CASES: list[tuple[str, str]] = [
    # Missing protocol (and not a recognized URL either)
    ("gpt2", "Must start with 'hf://'"),
    ("hf:/gpt2", "Must start with 'hf://'"),
    ("/local/path/file.txt", "Must start with 'hf://'"),
    # Empty body after protocol
    ("hf://", "Empty body"),
    # Empty repo id (just a type prefix)
    ("hf://datasets", "Missing identifier"),
    ("hf://datasets/", "Missing repository id"),
    ("hf://models/", "Missing repository id"),
    ("hf://buckets", "Missing identifier"),
    ("hf://buckets/", "Bucket id must be 'namespace/name'"),
    # Singular type forms are forbidden
    ("hf://dataset/foo/bar", "Type prefix must be plural"),
    ("hf://model/my-org/my-model", "Type prefix must be plural"),
    ("hf://space/user/my-space", "Type prefix must be plural"),
    ("hf://bucket/org/b", "Type prefix must be plural"),
    # Canonical repos (no namespace) are forbidden
    ("hf://gpt2", "Repository id must be 'namespace/name'"),
    ("hf://models/gpt2", "Repository id must be 'namespace/name'"),
    ("hf://datasets/squad", "Repository id must be 'namespace/name'"),
    ("hf://gpt2@v1", "Repository id must be 'namespace/name'"),
    ("hf://gpt2@v1/config.json", "Repository id must be 'namespace/name'"),
    # Buckets must always have namespace/name
    ("hf://buckets/single-segment", "Bucket id must be 'namespace/name'"),
    # Buckets cannot have a revision
    ("hf://buckets/org/b@v1", "do not support a revision"),
    ("hf://buckets/org/b@v1/path", "do not support a revision"),
    # Empty revision
    ("hf://my-org/my-model@", "Empty revision"),
    ("hf://datasets/foo/bar@/file", "Empty revision"),
    # Empty repo id before '@'
    ("hf://@v1/file", "Missing repository id"),
    # Invalid repo id chars (validated by validate_repo_id)
    ("hf://datasets/foo/.invalid", "Repo id must use alphanumeric"),
    ("hf://models/foo--bar/baz", "Cannot have -- or .."),
    # Empty path segments (adjacent slashes)
    ("hf://models/org/m//sub", "empty segments"),
    ("hf://buckets/org/b//sub", "empty segments"),
    ("hf://models/org/m/sub//dir", "empty segments"),
    ("hf://models/org/m@main//file.txt", "empty segments"),
    ("hf://datasets/foo/bar@refs/pr/10//file.csv", "empty segments"),
]


# ---------------------------------------------------------------------------
# HfUri direct init invalid cases: (kwargs, error_substring)
# ---------------------------------------------------------------------------
URI_DIRECT_INIT_INVALID_CASES: list[tuple[dict, str]] = [
    ({"type": "unknown", "id": "org/repo"}, "Invalid type"),
    ({"type": "model", "id": "gpt2"}, "namespace/name"),
    ({"type": "model", "id": "a/b/c"}, "namespace/name"),
    ({"type": "dataset", "id": "foo--bar/baz"}, "Cannot have -- or .."),
    ({"type": "model", "id": "org/repo", "revision": ""}, "empty string"),
    ({"type": "bucket", "id": "org/bucket", "revision": "main"}, "do not support a revision"),
    ({"type": "model", "id": "org/repo", "path_in_repo": "a//b"}, "empty segments"),
]


# ---------------------------------------------------------------------------
# URL parsing success cases: (url, expected_HfUri)
#
# These are Hugging Face *web URLs* (copy-pasted from the browser) that
# 'parse_hf_uri' normalizes to the canonical 'hf://' form.
# ---------------------------------------------------------------------------
URL_SUCCESS_CASES: list[tuple[str, HfUri]] = [
    # --- Repository landing pages ---------------------------------------------
    (
        "https://huggingface.co/my-org/my-model",
        HfUri(type="model", id="my-org/my-model"),
    ),
    (
        "https://huggingface.co/datasets/my-org/my-dataset",
        HfUri(type="dataset", id="my-org/my-dataset"),
    ),
    (
        "https://huggingface.co/spaces/user/my-space",
        HfUri(type="space", id="user/my-space"),
    ),
    (
        "https://huggingface.co/kernels/my-org/my-kernel",
        HfUri(type="kernel", id="my-org/my-kernel"),
    ),
    # Explicit 'models/' prefix is accepted too.
    (
        "https://huggingface.co/models/my-org/my-model",
        HfUri(type="model", id="my-org/my-model"),
    ),
    # Trailing slash is tolerated.
    (
        "https://huggingface.co/my-org/my-model/",
        HfUri(type="model", id="my-org/my-model"),
    ),
    # --- File viewer routes (blob / resolve / raw / blame) ---------------------
    (
        "https://huggingface.co/my-org/my-model/blob/main/config.json",
        HfUri(type="model", id="my-org/my-model", revision="main", path_in_repo="config.json"),
    ),
    (
        "https://huggingface.co/datasets/my-org/my-dataset/resolve/v1/data/train.csv",
        HfUri(type="dataset", id="my-org/my-dataset", revision="v1", path_in_repo="data/train.csv"),
    ),
    (
        "https://huggingface.co/my-org/my-model/raw/main/README.md",
        HfUri(type="model", id="my-org/my-model", revision="main", path_in_repo="README.md"),
    ),
    (
        "https://huggingface.co/spaces/user/my-space/blame/main/app.py",
        HfUri(type="space", id="user/my-space", revision="main", path_in_repo="app.py"),
    ),
    # --- Folder viewer route (tree) -------------------------------------------
    (
        "https://huggingface.co/my-org/my-model/tree/main/subdir",
        HfUri(type="model", id="my-org/my-model", revision="main", path_in_repo="subdir"),
    ),
    # '/tree/<rev>' with no path -> repo root pinned at a revision.
    (
        "https://huggingface.co/my-org/my-model/tree/v2.0",
        HfUri(type="model", id="my-org/my-model", revision="v2.0"),
    ),
    # --- Special revisions -----------------------------------------------------
    (
        "https://huggingface.co/datasets/foo/bar/blob/refs/pr/10/file.csv",
        HfUri(type="dataset", id="foo/bar", revision="refs/pr/10", path_in_repo="file.csv"),
    ),
    (
        "https://huggingface.co/datasets/foo/bar/resolve/refs/convert/parquet/data.parquet",
        HfUri(type="dataset", id="foo/bar", revision="refs/convert/parquet", path_in_repo="data.parquet"),
    ),
    # URL-encoded special revision.
    (
        "https://huggingface.co/datasets/foo/bar/blob/refs%2Fpr%2F10/file.csv",
        HfUri(type="dataset", id="foo/bar", revision="refs/pr/10", path_in_repo="file.csv"),
    ),
    # URL-encoded branch name containing '/'.
    (
        "https://huggingface.co/my-org/my-model/blob/feature%2Ffoo/config.json",
        HfUri(type="model", id="my-org/my-model", revision="feature/foo", path_in_repo="config.json"),
    ),
    # --- Percent-encoded path (browsers encode spaces, '#', ... in file names) --
    (
        "https://huggingface.co/my-org/my-model/blob/main/my%20file.txt",
        HfUri(type="model", id="my-org/my-model", revision="main", path_in_repo="my file.txt"),
    ),
    (
        "https://huggingface.co/my-org/my-model/blob/main/dir/a%23b.txt",
        HfUri(type="model", id="my-org/my-model", revision="main", path_in_repo="dir/a#b.txt"),
    ),
    # --- Buckets (no revision; '/resolve/' and '/tree/' are followed by the path) ---
    (
        "https://huggingface.co/buckets/my-org/my-bucket",
        HfUri(type="bucket", id="my-org/my-bucket"),
    ),
    (
        "https://huggingface.co/buckets/my-org/my-bucket/resolve/sub/data.bin",
        HfUri(type="bucket", id="my-org/my-bucket", path_in_repo="sub/data.bin"),
    ),
    (
        "https://huggingface.co/buckets/my-org/my-bucket/tree/sub/dir",
        HfUri(type="bucket", id="my-org/my-bucket", path_in_repo="sub/dir"),
    ),
    # Percent-encoded bucket path.
    (
        "https://huggingface.co/buckets/my-org/my-bucket/resolve/sub/my%20data.bin",
        HfUri(type="bucket", id="my-org/my-bucket", path_in_repo="sub/my data.bin"),
    ),
    # --- Query string and fragment are dropped --------------------------------
    (
        "https://huggingface.co/datasets/my-org/my-dataset/blob/main/train.csv?download=true",
        HfUri(type="dataset", id="my-org/my-dataset", revision="main", path_in_repo="train.csv"),
    ),
    (
        "https://huggingface.co/my-org/my-model/blob/main/config.json#L10",
        HfUri(type="model", id="my-org/my-model", revision="main", path_in_repo="config.json"),
    ),
    # --- Alternative hosts / scheme-less --------------------------------------
    (
        "https://hf.co/my-org/my-model",
        HfUri(type="model", id="my-org/my-model"),
    ),
    (
        "http://huggingface.co/my-org/my-model",
        HfUri(type="model", id="my-org/my-model"),
    ),
    (
        "huggingface.co/datasets/my-org/my-dataset",
        HfUri(type="dataset", id="my-org/my-dataset"),
    ),
]


# ---------------------------------------------------------------------------
# URL parsing failure cases: (url, error_substring)
# ---------------------------------------------------------------------------
URL_FAILURE_CASES: list[tuple[str, str]] = [
    # Unrecognized host
    ("https://example.com/my-org/my-model", "Unrecognized host"),
    ("https://huggingface.co.evil.com/my-org/my-model", "Unrecognized host"),
    # Single-segment URLs (user/org profile, canonical repo, listing pages) are ambiguous
    ("https://huggingface.co/gpt2", "namespace"),
    ("https://huggingface.co/datasets", "namespace"),
    ("https://huggingface.co/settings", "namespace"),
    # Unsupported repo routes (not a file/folder location)
    ("https://huggingface.co/my-org/my-model/commit/abc123", "unsupported"),
    ("https://huggingface.co/my-org/my-model/commits/main", "unsupported"),
    ("https://huggingface.co/my-org/my-model/discussions", "unsupported"),
    ("https://huggingface.co/my-org/my-model/settings", "unsupported"),
    # '/media/' is not a route we map to a Hub location.
    ("https://huggingface.co/my-org/my-model/media/main/image.png", "unsupported"),
    # Collections, blog, docs, ... are not repositories
    ("https://huggingface.co/collections/my-org/my-collection-abc123", "unsupported"),
    # Canonical repo (no namespace): 'squad/blob' is read as the id, so the next segment is
    # an unrecognized route -> rejected (canonical repos are not supported anyway).
    ("https://huggingface.co/datasets/squad/blob/main/file.json", "unsupported"),
    # Unsupported bucket route
    ("https://huggingface.co/buckets/my-org/my-bucket/settings", "unsupported"),
    # Empty body
    ("https://huggingface.co/", "Missing repository or bucket identifier"),
    ("https://huggingface.co", "Missing repository or bucket identifier"),
]


# ---------------------------------------------------------------------------
# 'HfUri.to_url' cases: (HfUri, expected_url_path)
#
# 'expected_url_path' is everything after the endpoint; the endpoint is prepended
# at test time so the cases stay independent of staging vs production.
# ---------------------------------------------------------------------------
TO_URL_CASES: list[tuple[HfUri, str]] = [
    # Repository landing pages
    (
        HfUri(type="model", id="my-org/my-model"),
        "/my-org/my-model",
    ),
    (
        HfUri(type="dataset", id="my-org/my-dataset"),
        "/datasets/my-org/my-dataset",
    ),
    (
        HfUri(type="space", id="user/my-space"),
        "/spaces/user/my-space",
    ),
    (
        HfUri(type="kernel", id="my-org/my-kernel"),
        "/kernels/my-org/my-kernel",
    ),
    # Revision only -> folder viewer at that revision
    (
        HfUri(type="model", id="my-org/my-model", revision="v1.0"),
        "/my-org/my-model/tree/v1.0",
    ),
    # File with explicit revision -> blob
    (
        HfUri(type="dataset", id="my-org/my-dataset", revision="v1", path_in_repo="train.csv"),
        "/datasets/my-org/my-dataset/blob/v1/train.csv",
    ),
    # File without revision -> blob at 'main'
    (
        HfUri(type="model", id="my-org/my-model", path_in_repo="config.json"),
        "/my-org/my-model/blob/main/config.json",
    ),
    # Special ref is kept verbatim
    (
        HfUri(type="dataset", id="foo/bar", revision="refs/pr/10", path_in_repo="file.csv"),
        "/datasets/foo/bar/blob/refs/pr/10/file.csv",
    ),
    # Branch name with '/' is URL-encoded so it stays a single segment
    (
        HfUri(type="model", id="my-org/my-model", revision="feature/foo", path_in_repo="config.json"),
        "/my-org/my-model/blob/feature%2Ffoo/config.json",
    ),
    # Branch name with '#' is percent-encoded so it is not read as a URL fragment (blob route)
    (
        HfUri(type="model", id="my-org/my-model", revision="fix#1", path_in_repo="config.json"),
        "/my-org/my-model/blob/fix%231/config.json",
    ),
    # '#' in the revision is also encoded on the folder (tree) route
    (
        HfUri(type="dataset", id="my-org/my-dataset", revision="fix#1"),
        "/datasets/my-org/my-dataset/tree/fix%231",
    ),
    # Path with special characters is percent-encoded ('/' kept as the separator)
    (
        HfUri(type="model", id="my-org/my-model", revision="main", path_in_repo="dir/my file#1.txt"),
        "/my-org/my-model/blob/main/dir/my%20file%231.txt",
    ),
    # Buckets: landing page and tree route (no revision)
    (
        HfUri(type="bucket", id="my-org/my-bucket"),
        "/buckets/my-org/my-bucket",
    ),
    (
        HfUri(type="bucket", id="my-org/my-bucket", path_in_repo="sub/data.bin"),
        "/buckets/my-org/my-bucket/tree/sub/data.bin",
    ),
]


# ---------------------------------------------------------------------------
# HfMount success cases: (mount_str, expected_HfMount, expected_roundtrip)
# ---------------------------------------------------------------------------
MOUNT_SUCCESS_CASES: list[tuple[str, HfMount, str]] = [
    (
        "hf://my-org/my-model:/data",
        HfMount(source=HfUri(type="model", id="my-org/my-model"), mount_path="/data"),
        "hf://models/my-org/my-model:/data",
    ),
    (
        "hf://my-org/my-model:/data:ro",
        HfMount(source=HfUri(type="model", id="my-org/my-model"), mount_path="/data", read_only=True),
        "hf://models/my-org/my-model:/data:ro",
    ),
    (
        "hf://my-org/my-model:/data:rw",
        HfMount(source=HfUri(type="model", id="my-org/my-model"), mount_path="/data", read_only=False),
        "hf://models/my-org/my-model:/data:rw",
    ),
    (
        "hf://datasets/my-org/my-dataset:/mnt",
        HfMount(source=HfUri(type="dataset", id="my-org/my-dataset"), mount_path="/mnt"),
        "hf://datasets/my-org/my-dataset:/mnt",
    ),
    # Mount path with revision
    (
        "hf://datasets/my-org/my-dataset@v1:/mnt:ro",
        HfMount(
            source=HfUri(type="dataset", id="my-org/my-dataset", revision="v1"),
            mount_path="/mnt",
            read_only=True,
        ),
        "hf://datasets/my-org/my-dataset@v1:/mnt:ro",
    ),
    # Mount path with sub-path inside repo
    (
        "hf://datasets/my-org/my-dataset/train:/mnt",
        HfMount(
            source=HfUri(type="dataset", id="my-org/my-dataset", path_in_repo="train"),
            mount_path="/mnt",
        ),
        "hf://datasets/my-org/my-dataset/train:/mnt",
    ),
    # Bucket with mount path and ro/rw
    (
        "hf://buckets/my-org/my-bucket:/mnt:rw",
        HfMount(
            source=HfUri(type="bucket", id="my-org/my-bucket"),
            mount_path="/mnt",
            read_only=False,
        ),
        "hf://buckets/my-org/my-bucket:/mnt:rw",
    ),
    # Bucket with sub-path and mount
    (
        "hf://buckets/my-org/my-bucket/sub/dir:/mnt:ro",
        HfMount(
            source=HfUri(type="bucket", id="my-org/my-bucket", path_in_repo="sub/dir"),
            mount_path="/mnt",
            read_only=True,
        ),
        "hf://buckets/my-org/my-bucket/sub/dir:/mnt:ro",
    ),
    # Mount path with several path segments
    (
        "hf://my-org/my-model:/path/to/mount",
        HfMount(source=HfUri(type="model", id="my-org/my-model"), mount_path="/path/to/mount"),
        "hf://models/my-org/my-model:/path/to/mount",
    ),
]


# ---------------------------------------------------------------------------
# HfMount failure cases: (mount_str, error_substring)
# ---------------------------------------------------------------------------
MOUNT_FAILURE_CASES: list[tuple[str, str]] = [
    # Missing protocol
    ("my-org/my-model:/data", "Must start with 'hf://'"),
    # Missing mount path entirely
    ("hf://my-org/my-model", "Missing mount path"),
    # Mount path that is just '/'
    ("hf://my-org/my-model:/", "Mount path must be a non-empty absolute path"),
    # Read-only flag without a mount path
    ("hf://my-org/my-model:ro", "':ro'/':rw' suffix is only valid"),
    ("hf://my-org/my-model:rw", "':ro'/':rw' suffix is only valid"),
    # Invalid URI inside the mount (canonical repo without namespace)
    ("hf://gpt2:/data", "Repository id must be 'namespace/name'"),
    ("hf://gpt2:/data:ro", "Repository id must be 'namespace/name'"),
]


# ---------------------------------------------------------------------------
# HfMount direct init invalid cases: (kwargs, error_substring)
# ---------------------------------------------------------------------------
MOUNT_DIRECT_INIT_INVALID_CASES: list[tuple[dict, str]] = [
    ({"source": HfUri(type="model", id="org/repo"), "mount_path": "relative"}, "absolute path"),
    ({"source": HfUri(type="model", id="org/repo"), "mount_path": "/"}, "absolute path"),
]


# ===========================================================================
# Tests
# ===========================================================================


@pytest.mark.parametrize(("uri", "expected", "expected_roundtrip"), URI_SUCCESS_CASES)
def test_parse_hf_uri_success(uri: str, expected: HfUri, expected_roundtrip: str) -> None:
    result = parse_hf_uri(uri)
    assert result == expected
    assert result.to_uri() == expected_roundtrip
    # Re-parsing the canonical form must yield the same URI (idempotency).
    assert parse_hf_uri(expected_roundtrip) == expected


@pytest.mark.parametrize(("uri", "error_substring"), URI_FAILURE_CASES)
def test_parse_hf_uri_failure(uri: str, error_substring: str) -> None:
    with pytest.raises(HfUriError, match=error_substring) as exc_info:
        parse_hf_uri(uri)
    assert exc_info.value.uri == uri


@pytest.mark.parametrize(("kwargs", "error_substring"), URI_DIRECT_INIT_INVALID_CASES)
def test_hf_uri_direct_init_invalid(kwargs: dict, error_substring: str) -> None:
    with pytest.raises(HfUriError, match=error_substring):
        HfUri(**kwargs)


@pytest.mark.parametrize(("url", "expected"), URL_SUCCESS_CASES)
def test_parse_hf_uri_from_url_success(url: str, expected: HfUri) -> None:
    assert parse_hf_uri(url) == expected


@pytest.mark.parametrize(("url", "error_substring"), URL_FAILURE_CASES)
def test_parse_hf_uri_from_url_failure(url: str, error_substring: str) -> None:
    with pytest.raises(HfUriError, match=error_substring) as exc_info:
        parse_hf_uri(url)
    assert exc_info.value.uri == url


def test_parse_hf_uri_custom_endpoint() -> None:
    # URLs on a custom Hub endpoint are recognized when 'endpoint' is passed...
    assert parse_hf_uri(
        "https://hub.my-company.com/datasets/my-org/my-dataset/blob/main/train.csv",
        endpoint="https://hub.my-company.com",
    ) == HfUri(type="dataset", id="my-org/my-dataset", revision="main", path_in_repo="train.csv")

    # ... including self-hosted endpoints that carry a path prefix (stripped before parsing).
    assert parse_hf_uri(
        "http://localhost:8080/hf/my-org/my-model",
        endpoint="http://localhost:8080/hf",
    ) == HfUri(type="model", id="my-org/my-model", revision=None, path_in_repo="")

    # ... but the same URL is rejected without the matching 'endpoint'.
    with pytest.raises(HfUriError, match="Unrecognized host"):
        parse_hf_uri("https://hub.my-company.com/my-org/my-model")


@pytest.mark.parametrize(("uri", "expected_path"), TO_URL_CASES)
def test_hf_uri_to_url(uri: HfUri, expected_path: str) -> None:
    expected = constants.ENDPOINT + expected_path
    assert uri.to_url() == expected
    # The rendered URL must parse back to a URI pointing at the same resource.
    reparsed = parse_hf_uri(expected)
    assert reparsed.type == uri.type
    assert reparsed.id == uri.id
    assert reparsed.path_in_repo == uri.path_in_repo
    # Revision round-trips exactly, except that a file URL without a revision defaults to 'main'.
    if uri.revision is not None:
        assert reparsed.revision == uri.revision


def test_hf_uri_to_url_custom_endpoint() -> None:
    uri = HfUri(type="dataset", id="my-org/my-dataset", revision="v1", path_in_repo="train.csv")
    assert uri.to_url(endpoint="https://hub-ci.huggingface.co") == (
        "https://hub-ci.huggingface.co/datasets/my-org/my-dataset/blob/v1/train.csv"
    )


@pytest.mark.parametrize(("mount_str", "expected", "expected_roundtrip"), MOUNT_SUCCESS_CASES)
def test_parse_hf_mount_success(mount_str: str, expected: HfMount, expected_roundtrip: str) -> None:
    result = parse_hf_mount(mount_str)
    assert result == expected
    assert result.to_uri() == expected_roundtrip
    # Re-parsing the canonical form must yield the same mount (idempotency).
    assert parse_hf_mount(expected_roundtrip) == expected


@pytest.mark.parametrize(("mount_str", "error_substring"), MOUNT_FAILURE_CASES)
def test_parse_hf_mount_failure(mount_str: str, error_substring: str) -> None:
    with pytest.raises(HfUriError, match=error_substring) as exc_info:
        parse_hf_mount(mount_str)
    assert exc_info.value.uri == mount_str


@pytest.mark.parametrize(("kwargs", "error_substring"), MOUNT_DIRECT_INIT_INVALID_CASES)
def test_hf_mount_direct_init_invalid(kwargs: dict, error_substring: str) -> None:
    with pytest.raises(HfUriError, match=error_substring):
        HfMount(**kwargs)



#!/usr/bin/env python3
"""Generate a Slack announcement message from existing release notes.

This script:
1. Reads the generated release notes markdown
2. Invokes an OpenCode skill to summarize into a Slack-formatted message
3. Appends a deterministic "Pinging" section with CI links
4. Outputs the final message to .release-notes/

Usage:
  python -m utils.release_notes.generate_slack_message --version v1.7.0 --rc-version 1.7.0rc0
"""

import argparse
import os
import shutil
import subprocess
import sys
from pathlib import Path

from .validate_notes import find_release_notes_file


OUTPUT_DIR = Path(os.environ.get("RELEASE_NOTES_OUTPUT_DIR", ".release-notes"))

# Downstream repos to ping for RC testing.
# Each entry: (display_label, slack_mention, repo_name_or_None)
# repo_name=None means no CI link (e.g., lighteval has no automated HF Hub testing branch)
DEFAULT_PING_LIST: list[tuple[str, str, str | None]] = [
    ("transformers", "@U01JNPUN1ML", "transformers"),  # Yih-Dar
    ("datasets", "@U011YKS85FY", "datasets"),  # Quentin
    ("diffusers", "@U03AU4E7DJB", "diffusers"),  # Sayak
    ("lighteval", "@U04MZDFL8DD", None),  # Nathan
    ("sentence-transformers", "@U04E4DNPWG7", "sentence-transformers"),  # Tom
]


def build_ping_section(
    ping_list: list[tuple[str, str, str | None]] | None = None,
) -> str:
    """Build the "Pinging:" block with CI compare URLs and closing line.

    Args:
        ping_list: List of (label, slack_mention, repo_name_or_None). Defaults to DEFAULT_PING_LIST.

    Returns:
        The pinging section + closing line as a string.
    """
    if ping_list is None:
        ping_list = DEFAULT_PING_LIST

    lines = ["Pinging:"]
    for label, mention, repo in ping_list:
        if repo is not None:
            lines.append(f"- for {label} {mention} => link")
        else:
            lines.append(f"- for {label} {mention}")

    lines.append("")
    lines.append("Let us know if you spot any regressions! Release should be happening pretty soon :hugging_face:")

    return "\n".join(lines)


def run_opencode_skill(
    version: str,
    rc_version: str,
    input_file: Path,
    output_file: Path,
) -> bool:
    """Run the hf-release-notes:slack OpenCode skill.

    Args:
        version: Base version (e.g., "v1.7.0")
        rc_version: RC version for pip install (e.g., "1.7.0rc0")
        input_file: Path to the release notes markdown
        output_file: Path to write the Slack message body

    Returns:
        True if successful, False otherwise
    """
    prompt = (
        f"Run the hf-release-notes:slack skill. "
        f"Generate a Slack announcement message for version {version}. "
        f"The RC version for the pip install command is {rc_version}. "
        f"Read the release notes from {input_file}. "
        f"Write the message body to {output_file}. "
        f"Do NOT include the Pinging section or closing line — those are appended by the script."
    )

    opencode_cmd = shutil.which("opencode")
    if not opencode_cmd:
        print("Error: 'opencode' command not found in PATH", file=sys.stderr)
        return False

    cmd = [opencode_cmd]
    model = os.environ.get("RELEASE_NOTES_MODEL")
    if model:
        cmd.extend(["--model", model])
    cmd.extend(["run", prompt])
    print(f"Running: {' '.join(cmd)}")

    try:
        result = subprocess.run(cmd, check=True, capture_output=False)
        return result.returncode == 0
    except subprocess.CalledProcessError as e:
        print(f"OpenCode command failed with exit code {e.returncode}", file=sys.stderr)
        return False
    except FileNotFoundError:
        print("Error: 'opencode' command not found", file=sys.stderr)
        return False


def main(version: str, rc_version: str, input_file: Path | None = None, output_file: Path | None = None) -> int:
    """Generate a Slack announcement message.

    Args:
        version: Base version (e.g., "v1.7.0")
        rc_version: RC version for pip install (e.g., "1.7.0rc0")
        input_file: Path to release notes (auto-detected if None)
        output_file: Output path (defaults to .release-notes/SLACK_MESSAGE_{version}.md)

    Returns:
        Exit code (0 for success, non-zero for failure)
    """
    # Resolve input file
    if input_file is None:
        input_file = find_release_notes_file(version)
    if input_file is None or not input_file.exists():
        print(f"Error: Release notes file not found for {version}", file=sys.stderr)
        print(f"Expected at: {OUTPUT_DIR}/RELEASE_NOTES_{version}.md", file=sys.stderr)
        return 1

    # Resolve output file
    if output_file is None:
        output_file = OUTPUT_DIR / f"SLACK_MESSAGE_{version}.md"

    # Clean previous output
    if output_file.exists():
        output_file.unlink()

    print(f"Input:  {input_file}")
    print(f"Output: {output_file}")

    # Generate message body with OpenCode
    body_file = OUTPUT_DIR / f"_slack_body_{version}.md"
    if not run_opencode_skill(version, rc_version, input_file, body_file):
        print("Failed to generate Slack message body", file=sys.stderr)
        return 1

    if not body_file.exists():
        print(f"Error: OpenCode did not produce output at {body_file}", file=sys.stderr)
        return 1

    body = body_file.read_text().strip()

    # Build pinging section
    ping_section = build_ping_section()

    # Combine and write final message
    final_message = f"{body}\n\n{ping_section}\n"
    output_file.write_text(final_message)

    # Clean up intermediate file
    body_file.unlink(missing_ok=True)

    # Print to stdout for convenience
    print("\n" + "=" * 60)
    print("SLACK MESSAGE")
    print("=" * 60)
    print(final_message)
    print("=" * 60)
    print(f"Saved to {output_file}")

    return 0


def cli():
    """CLI entry point."""
    parser = argparse.ArgumentParser(
        description="Generate a Slack announcement message from release notes",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Generate Slack message for v1.7.0 prerelease
  python -m utils.release_notes.generate_slack_message --version v1.7.0 --rc-version 1.7.0rc0

  # With custom input/output paths
  python -m utils.release_notes.generate_slack_message \\
    --version v1.7.0 --rc-version 1.7.0rc0 \\
    --input .release-notes/RELEASE_NOTES_v1.7.0.md \\
    --output slack_message.md
""",
    )
    parser.add_argument(
        "--version",
        required=True,
        help="Base version for release notes lookup (e.g., v1.7.0)",
    )
    parser.add_argument(
        "--rc-version",
        required=True,
        help="Full RC version for pip install command (e.g., 1.7.0rc0)",
    )
    parser.add_argument(
        "--input",
        type=Path,
        default=None,
        help="Path to release notes file (auto-detected if omitted)",
    )
    parser.add_argument(
        "--output",
        type=Path,
        default=None,
        help="Output path (defaults to .release-notes/SLACK_MESSAGE_{version}.md)",
    )
    args = parser.parse_args()

    sys.exit(main(args.version, args.rc_version, args.input, args.output))


if __name__ == "__main__":
    cli()



import time
from pathlib import Path

import pytest

from huggingface_hub.constants import ENDPOINT
from huggingface_hub.utils import run_interactive_subprocess, run_subprocess
from huggingface_hub.utils._git_credential import (
    _parse_credential_output,
    list_credential_helpers,
    set_git_credential,
    unset_git_credential,
)


STORE_AND_CACHE_HELPERS_CONFIG = """
[credential]
    helper = store
    helper = cache --timeout 30000
    helper = git-credential-manager
    helper = /usr/libexec/git-core/git-credential-libsecret
"""


class TestGitCredentials:
    @pytest.fixture(autouse=True)
    def setup(self, tmp_path):
        """Initialize and configure a local repo.

        Avoid to configure git helpers globally on a contributor's machine.
        """
        run_subprocess("git init", folder=tmp_path)
        with (tmp_path / ".git" / "config").open("w") as f:
            f.write(STORE_AND_CACHE_HELPERS_CONFIG)
        yield

    def test_list_credential_helpers(self, tmp_path: Path) -> None:
        helpers = list_credential_helpers(folder=tmp_path)
        assert "cache" in helpers
        assert "store" in helpers
        assert "git-credential-manager" in helpers
        assert "/usr/libexec/git-core/git-credential-libsecret" in helpers

    def test_set_and_unset_git_credential(self, tmp_path: Path) -> None:
        username = "hf_test_user_" + str(round(time.time()))  # make username unique

        # Set credentials
        set_git_credential(token="hf_test_token", username=username, folder=tmp_path)

        # Check credentials are stored
        with run_interactive_subprocess("git credential fill", folder=tmp_path) as (stdin, stdout):
            stdin.write(f"url={ENDPOINT}\nusername={username}\n\n")
            stdin.flush()
            output = stdout.read()
        assert "password=hf_test_token" in output

        # Unset credentials
        unset_git_credential(username=username, folder=tmp_path)

        # Check credentials are NOT stored
        # Cannot check with `git credential fill` as it would hang forever: only
        # checking `store` helper instead.
        with run_interactive_subprocess("git credential-store get", folder=tmp_path) as (stdin, stdout):
            stdin.write(f"url={ENDPOINT}\nusername={username}\n\n")
            stdin.flush()
            output = stdout.read()
        assert output == ""

    def test_git_credential_parsing_regex(self) -> None:
        output = """
            credential.helper = store
            credential.helper = cache --timeout 30000
        credential.helper = osxkeychain"""
        assert _parse_credential_output(output) == ["cache", "osxkeychain", "store"]



"""Unit tests for the streamed multi-commit upload pipeline (`_upload_pipeline.py`).

Everything is mocked: no network, no `hf_xet` runtime. The fake Xet session mimics the
`XetSession.new_upload_commit` contract (background uploads, sha256 computation, finalize).
"""

import hashlib
import json
import threading
from types import SimpleNamespace
from unittest.mock import patch

import pytest

import huggingface_hub._commit_api as commit_api
import huggingface_hub._upload_pipeline as upload_pipeline
from huggingface_hub._commit_api import CommitOperationAdd, CommitOperationDelete, _compute_missing_sha256s
from huggingface_hub._upload_pipeline import _CommitPacer, _LiveDisplay, _UploadPipeline


pytestmark = pytest.mark.xet

# ---------------------------------------------------------------------------
# Fakes
# ---------------------------------------------------------------------------


class FakeHandle:
    def __init__(self, sha256_hex: str):
        self._sha256 = sha256_hex

    def result(self):
        return SimpleNamespace(xet_info=SimpleNamespace(sha256=self._sha256))


class FakeXetCommit:
    """Mimics XetUploadCommit: computes sha256 'in the background' like COMPUTE_SHA256 does."""

    def __init__(self):
        self.files: list[str] = []
        self.finished = False
        self.aborted = False

    def start_upload_file(self, path, sha256=None):
        self.files.append(path)
        with open(path, "rb") as f:
            return FakeHandle(hashlib.sha256(f.read()).hexdigest())

    def start_upload_bytes(self, data, sha256=None, name=None):
        self.files.append(name)
        return FakeHandle(hashlib.sha256(data).hexdigest())

    def wait_to_finish(self):
        self.finished = True

    def abort(self):
        self.aborted = True


class FakeXetSession:
    def __init__(self):
        self.commits: list[FakeXetCommit] = []

    def new_upload_commit(self, **kwargs):
        commit = FakeXetCommit()
        self.commits.append(commit)
        return commit


class FakeCommitEndpoint:
    """Records commit POSTs and returns fake commit responses. Can fail on demand."""

    def __init__(self, fail_on_nth_call: set | None = None):
        self.calls: list[dict] = []  # {url, params, operations}
        self.fail_on = fail_on_nth_call or set()
        self.lock = threading.Lock()

    def __call__(self, method, url, headers=None, content=None, params=None):
        with self.lock:
            n = len(self.calls)
            payload = [json.loads(line) for line in content.decode().strip().split("\n")]
            self.calls.append({"url": url, "params": params, "payload": payload})
            if n in self.fail_on:
                raise RuntimeError(f"injected commit failure (call {n})")
            return SimpleNamespace(
                json=lambda: {
                    "commitUrl": f"https://hub.fake/user/repo/commit/{n}",
                    "commitOid": f"oid{n}",
                }
            )

    def committed_paths(self, call_idx):
        return [item["value"]["path"] for item in self.calls[call_idx]["payload"] if item["key"] != "header"]


def fake_fetch_upload_modes(upload_modes_config=None):
    """Returns a `_fetch_upload_modes` replacement. `.bin` files -> xet ("lfs" mode), others -> regular."""

    def _fake(additions, **kwargs):
        _fake.calls.append(kwargs)
        for op in additions:
            config = (upload_modes_config or {}).get(op.path_in_repo, {})
            op._upload_mode = config.get("mode", "lfs" if op.path_in_repo.endswith(".bin") else "regular")
            op._should_ignore = config.get("ignore", False)
            op._remote_oid = config.get("remote_oid")

    _fake.calls = []
    return _fake


@pytest.fixture
def fake_api(tmp_path):
    api = SimpleNamespace(
        endpoint="https://hub.fake",
        _build_hf_headers=lambda token=None: {"user-agent": "test"},
        repo_info=lambda **kwargs: SimpleNamespace(sha="latestsha"),
        pr_calls=[],
    )

    def create_pull_request(**kwargs):
        api.pr_calls.append(kwargs)
        return SimpleNamespace(url="https://huggingface.co/fake/repo/discussions/7", git_reference="refs/pr/7")

    api.create_pull_request = create_pull_request
    return api


def make_ops(tmp_path, names_and_content):
    ops = []
    for name, content in names_and_content:
        path = tmp_path / name.replace("/", "_")
        path.write_bytes(content)
        ops.append(CommitOperationAdd(path_in_repo=name, path_or_fileobj=str(path)))
    return ops


def run_pipeline(api, add_operations, commit_endpoint=None, modes=None, **kwargs):
    commit_endpoint = commit_endpoint or FakeCommitEndpoint()
    session = FakeXetSession()
    fetcher = fake_fetch_upload_modes(modes)
    commit_endpoint.preupload_calls = fetcher.calls  # exposed for assertions
    with (
        patch.object(upload_pipeline, "_fetch_upload_modes", fetcher),
        patch.object(upload_pipeline, "get_xet_session", lambda: session),
        patch.object(commit_api, "http_backoff", commit_endpoint),
        patch.object(commit_api, "hf_raise_for_status", lambda *a, **k: None),
        patch.object(upload_pipeline, "are_progress_bars_disabled", lambda: True),
    ):
        pipeline = _UploadPipeline(
            api,
            repo_id="user/repo",
            repo_type="model",
            add_operations=add_operations,
            delete_operations=kwargs.pop("delete_operations", []),
            commit_message=kwargs.pop("commit_message", "msg"),
            commit_description=None,
            token=None,
            revision=kwargs.pop("revision", None),
            create_pr=kwargs.pop("create_pr", False),
            parent_commit=kwargs.pop("parent_commit", None),
        )
        info = pipeline.run()
    return info, commit_endpoint, session


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------


class TestCommitPacer:
    def test_scale_up_on_fast_full_commit(self):
        pacer = _CommitPacer()
        initial = pacer.target
        pacer.record_success(duration=2.0, nb_files=initial)
        assert pacer.target > initial

    def test_no_scale_up_on_partial_commit(self):
        pacer = _CommitPacer()
        initial = pacer.target
        pacer.record_success(duration=2.0, nb_files=initial - 1)
        assert pacer.target == initial

    def test_scale_down_on_slow_commit(self):
        pacer = _CommitPacer()
        initial = pacer.target
        pacer.record_success(duration=100.0, nb_files=initial)
        assert pacer.target < initial

    def test_scale_down_on_failure_and_bounds(self):
        pacer = _CommitPacer()
        for _ in range(20):
            pacer.record_failure()
        assert pacer.target == upload_pipeline.COMMIT_SIZE_SCALE[0]
        for _ in range(20):
            pacer.record_success(duration=1.0, nb_files=pacer.target)
        assert pacer.target == upload_pipeline.COMMIT_SIZE_SCALE[-1]


class TestLiveDisplayCounters:
    def test_xet_callback_sums_increments_across_concurrent_commits(self):
        with patch.object(upload_pipeline.logger, "isEnabledFor", return_value=True):
            display = _LiveDisplay(total_files=10, enabled=True)
        cb1, cb2 = display.new_xet_callback(), display.new_xet_callback()

        def report(n):
            return SimpleNamespace(total_transfer_bytes_completed=n)

        cb1(report(100), {})
        cb2(report(50), {})  # concurrent commit with its own cumulative counter
        cb1(report(300), {})
        assert display._xet_bytes == 350

    def test_disabled_display_returns_no_callback_when_logging_off(self):
        with patch.object(upload_pipeline.logger, "isEnabledFor", return_value=False):
            display = _LiveDisplay(total_files=10, enabled=False)
        assert display.new_xet_callback() is None

    def test_disabled_display_keeps_callback_for_log_summaries(self):
        # Disabling progress bars (e.g. agent mode) must not kill the periodic log summaries.
        with patch.object(upload_pipeline.logger, "isEnabledFor", return_value=True):
            display = _LiveDisplay(total_files=10, enabled=False)
        assert display.new_xet_callback() is not None


class TestUploadPipeline:
    def test_single_commit_mixed_files(self, fake_api, tmp_path):
        ops = make_ops(tmp_path, [("a.txt", b"regular"), ("b.bin", b"x" * 1000), ("c.bin", b"y" * 1000)])
        info, endpoint, session = run_pipeline(fake_api, ops)

        assert len(endpoint.calls) == 1
        assert sorted(endpoint.committed_paths(0)) == ["a.txt", "b.bin", "c.bin"]
        # xet files uploaded through one upload-commit, finalized before the git commit
        assert len(session.commits) == 1
        assert session.commits[0].finished
        assert len(session.commits[0].files) == 2
        # sha256 backfilled from the xet result for unhashed operations
        assert ops[1].upload_info.sha256 == hashlib.sha256(b"x" * 1000).digest()
        assert all(op._is_committed for op in ops)
        assert info.oid == "oid0"

    def test_multi_commit_with_suffix_and_first_commit_extras(self, fake_api, tmp_path):
        ops = make_ops(tmp_path, [(f"f{i}.bin", f"{i}".encode() * 100) for i in range(5)])
        deletes = [CommitOperationDelete(path_in_repo="old.bin")]
        with (
            patch.object(upload_pipeline, "COMMIT_SIZE_SCALE", [2, 2, 2]),
            patch.object(upload_pipeline, "INITIAL_COMMIT_SIZE_INDEX", 0),
        ):
            info, endpoint, _ = run_pipeline(
                fake_api, ops, delete_operations=deletes, parent_commit="a" * 40, commit_message="msg"
            )

        assert len(endpoint.calls) == 3  # 2 + 2 + 1 files
        headers = [call["payload"][0]["value"] for call in endpoint.calls]
        assert headers[0]["summary"] == "msg"
        assert headers[1]["summary"] == "msg (part 2)"
        assert headers[2]["summary"] == "msg (part 3)"
        # deletions and parent_commit ride the first commit only
        assert headers[0].get("parentCommit") == "a" * 40
        assert "parentCommit" not in headers[1]
        assert "old.bin" in endpoint.committed_paths(0)
        assert info.oid == "oid2"  # last commit

    def test_unchanged_files_are_skipped(self, fake_api, tmp_path):
        content = b"z" * 1000
        sha = hashlib.sha256(content).hexdigest()
        ops = make_ops(tmp_path, [("same.bin", content), ("new.bin", b"w" * 1000)])
        modes = {"same.bin": {"remote_oid": sha}}
        info, endpoint, _ = run_pipeline(fake_api, ops, modes=modes)

        assert len(endpoint.calls) == 1
        assert endpoint.committed_paths(0) == ["new.bin"]

    def test_everything_unchanged_creates_no_commit(self, fake_api, tmp_path):
        content = b"z" * 1000
        ops = make_ops(tmp_path, [("same.bin", content)])
        modes = {"same.bin": {"remote_oid": hashlib.sha256(content).hexdigest()}}
        info, endpoint, _ = run_pipeline(fake_api, ops, modes=modes)

        assert len(endpoint.calls) == 0
        assert info.oid == "latestsha"  # from repo_info

    def test_gitignored_files_are_not_committed(self, fake_api, tmp_path):
        ops = make_ops(tmp_path, [("kept.bin", b"a" * 100), ("ignored.bin", b"b" * 100)])
        modes = {"ignored.bin": {"ignore": True}}
        _, endpoint, session = run_pipeline(fake_api, ops, modes=modes)

        assert endpoint.committed_paths(0) == ["kept.bin"]
        assert len(session.commits[0].files) == 1  # ignored file never registered with xet

    def test_create_pr_single_pr_for_multiple_commits(self, fake_api, tmp_path):
        ops = make_ops(tmp_path, [(f"f{i}.bin", f"{i}".encode() * 100) for i in range(4)])
        with (
            patch.object(upload_pipeline, "COMMIT_SIZE_SCALE", [2, 2]),
            patch.object(upload_pipeline, "INITIAL_COMMIT_SIZE_INDEX", 0),
            patch.object(upload_pipeline, "PREUPLOAD_BATCH_SIZE", 2),  # several preupload calls
        ):
            info, endpoint, _ = run_pipeline(fake_api, ops, create_pr=True)

        assert len(endpoint.calls) == 2
        # a single PR is created upfront (via the discussions API, not `?create_pr=1`)...
        assert len(fake_api.pr_calls) == 1
        assert fake_api.pr_calls[0]["title"] == "msg"
        # ...and ALL commits push to the (url-quoted) PR ref
        assert all(call["url"].endswith("/commit/refs%2Fpr%2F7") for call in endpoint.calls)
        assert all(not call["params"] for call in endpoint.calls)
        assert info.pr_url == "https://huggingface.co/fake/repo/discussions/7"
        assert info.pr_revision == "refs/pr/7"
        # preupload always targets the base revision with create_pr=True, even after the PR exists
        # (mirrors `create_commit` semantics; the PR ref is only used for the commit calls)
        assert all(call["revision"] == "main" and call["create_pr"] is True for call in endpoint.preupload_calls)

    def test_commit_failure_splits_batch(self, fake_api, tmp_path):
        ops = make_ops(tmp_path, [(f"f{i}.bin", f"{i}".encode() * 100) for i in range(4)])
        endpoint = FakeCommitEndpoint(fail_on_nth_call={0})  # first attempt (4 files) fails
        with (
            patch.object(upload_pipeline, "COMMIT_SIZE_SCALE", [2, 4]),
            patch.object(upload_pipeline, "INITIAL_COMMIT_SIZE_INDEX", 1),
        ):
            info, endpoint, _ = run_pipeline(fake_api, ops, commit_endpoint=endpoint)

        # 1 failed attempt + 2 successful halves
        assert len(endpoint.calls) == 3
        assert len(endpoint.committed_paths(1)) == 2
        assert len(endpoint.committed_paths(2)) == 2
        committed = endpoint.committed_paths(1) + endpoint.committed_paths(2)
        assert sorted(committed) == [f"f{i}.bin" for i in range(4)]

    def test_create_pr_interrupted_warns_with_resume_instructions(self, fake_api, tmp_path, caplog):
        """If the upload fails after the PR was created, tell the user how to resume into the SAME PR
        (re-running with create_pr=True would open a second one)."""
        ops = make_ops(tmp_path, [(f"f{i}.bin", f"{i}".encode() * 100) for i in range(2)])
        endpoint = FakeCommitEndpoint(fail_on_nth_call=set(range(100)))  # PR created, then every commit fails
        with (
            patch.object(upload_pipeline, "COMMIT_SIZE_SCALE", [1, 1]),
            patch.object(upload_pipeline, "INITIAL_COMMIT_SIZE_INDEX", 0),
        ):
            with pytest.raises(RuntimeError, match="injected commit failure"):
                run_pipeline(fake_api, ops, commit_endpoint=endpoint, create_pr=True)
        assert 'revision="refs/pr/7"' in caplog.text

    def test_create_pr_with_revision_raises(self, tmp_path):
        """PRs are always opened against the default branch (the discussions API has no base-revision
        support), so `create_pr=True` + `revision` is rejected upfront."""
        from huggingface_hub import HfApi

        with pytest.raises(ValueError, match="create_pr"):
            HfApi().upload_folder(repo_id="user/repo", folder_path=tmp_path, create_pr=True, revision="my-branch")

    def test_persistent_commit_failure_raises(self, fake_api, tmp_path):
        ops = make_ops(tmp_path, [("f.bin", b"x" * 100)])
        endpoint = FakeCommitEndpoint(fail_on_nth_call=set(range(100)))
        with pytest.raises(RuntimeError, match="injected commit failure"):
            run_pipeline(fake_api, ops, commit_endpoint=endpoint)

    def test_coordinator_error_propagates_without_hanging(self, fake_api, tmp_path):
        ops = make_ops(tmp_path, [("f.bin", b"x" * 100)])

        def failing_fetch(additions, **kwargs):
            raise ValueError("preupload exploded")

        session = FakeXetSession()
        with (
            patch.object(upload_pipeline, "_fetch_upload_modes", failing_fetch),
            patch.object(upload_pipeline, "get_xet_session", lambda: session),
            patch.object(upload_pipeline, "abort_xet_session", lambda: None),
            patch.object(upload_pipeline, "are_progress_bars_disabled", lambda: True),
        ):
            pipeline = _UploadPipeline(
                fake_api,
                repo_id="user/repo",
                repo_type="model",
                add_operations=ops,
                delete_operations=[],
                commit_message="msg",
                commit_description=None,
                token=None,
                revision=None,
                create_pr=False,
                parent_commit=None,
            )
            with pytest.raises(ValueError, match="preupload exploded"):
                pipeline.run()


class TestComputeMissingSha256s:
    def test_hashes_only_missing(self, tmp_path):
        path = tmp_path / "f.bin"
        path.write_bytes(b"content")
        op_lazy = CommitOperationAdd(path_in_repo="f.bin", path_or_fileobj=str(path))
        op_eager = CommitOperationAdd(path_in_repo="b.bin", path_or_fileobj=b"other")
        assert not op_lazy.upload_info.is_hashed
        assert op_eager.upload_info.is_hashed

        _compute_missing_sha256s([op_lazy, op_eager], num_threads=2)
        assert op_lazy.upload_info.is_hashed
        assert op_lazy.upload_info.sha256 == hashlib.sha256(b"content").digest()



import logging
from queue import Queue
from unittest.mock import Mock

import pytest

from huggingface_hub.utils._telemetry import send_telemetry

from .testing_constants import ENDPOINT_STAGING


class TestSendTelemetry:
    @pytest.fixture(autouse=True)
    def setup(self, mocker):
        self.queue: Queue = Queue()
        mocker.patch("huggingface_hub.utils._telemetry._TELEMETRY_QUEUE", self.queue)
        mocker.patch("huggingface_hub.utils._telemetry._TELEMETRY_THREAD", None)

        get_session_mock = Mock()
        self.mock_head = get_session_mock().head
        mocker.patch("huggingface_hub.utils._telemetry.get_session", get_session_mock)
        yield

    def test_topic_normal(self) -> None:
        send_telemetry(topic="examples")
        self.queue.join()  # Wait for the telemetry tasks to be completed
        self.mock_head.assert_called_once()
        assert self.mock_head.call_args[0][0] == f"{ENDPOINT_STAGING}/api/telemetry/examples"

    def test_topic_multiple(self) -> None:
        send_telemetry(topic="example1")
        send_telemetry(topic="example2")
        send_telemetry(topic="example3")
        self.queue.join()  # Wait for the telemetry tasks to be completed

        assert self.mock_head.call_count == 3  # 3 calls and order is preserved
        assert self.mock_head.call_args_list[0][0][0] == f"{ENDPOINT_STAGING}/api/telemetry/example1"
        assert self.mock_head.call_args_list[1][0][0] == f"{ENDPOINT_STAGING}/api/telemetry/example2"
        assert self.mock_head.call_args_list[2][0][0] == f"{ENDPOINT_STAGING}/api/telemetry/example3"

    def test_topic_with_subtopic(self) -> None:
        send_telemetry(topic="gradio/image/this_one")
        self.queue.join()  # Wait for the telemetry tasks to be completed
        self.mock_head.assert_called_once()
        assert self.mock_head.call_args[0][0] == f"{ENDPOINT_STAGING}/api/telemetry/gradio/image/this_one"

    def test_topic_quoted(self) -> None:
        send_telemetry(topic="foo bar")
        self.queue.join()  # Wait for the telemetry tasks to be completed
        self.mock_head.assert_called_once()
        assert self.mock_head.call_args[0][0] == f"{ENDPOINT_STAGING}/api/telemetry/foo%20bar"

    def test_hub_offline(self, mocker) -> None:
        mocker.patch("huggingface_hub.utils._telemetry.constants.HF_HUB_OFFLINE", True)
        send_telemetry(topic="topic")
        assert self.queue.empty()  # no tasks
        self.mock_head.assert_not_called()

    def test_telemetry_disabled(self, mocker) -> None:
        mocker.patch("huggingface_hub.utils._telemetry.constants.HF_HUB_DISABLE_TELEMETRY", True)
        send_telemetry(topic="topic")
        assert self.queue.empty()  # no tasks
        self.mock_head.assert_not_called()

    def test_telemetry_use_build_hf_headers(self, mocker) -> None:
        mock_headers = mocker.patch("huggingface_hub.utils._telemetry.build_hf_headers")
        send_telemetry(topic="topic")
        self.queue.join()  # Wait for the telemetry tasks to be completed
        self.mock_head.assert_called_once()
        mock_headers.assert_called_once()
        assert self.mock_head.call_args[1]["headers"] == mock_headers.return_value


class TestSendTelemetryConnectionError:
    @pytest.fixture(autouse=True)
    def setup(self, mocker):
        self.queue: Queue = Queue()
        mocker.patch("huggingface_hub.utils._telemetry._TELEMETRY_QUEUE", self.queue)
        mocker.patch("huggingface_hub.utils._telemetry._TELEMETRY_THREAD", None)

        get_session_mock = Mock()
        get_session_mock().head.side_effect = Exception("whatever")
        mocker.patch("huggingface_hub.utils._telemetry.get_session", get_session_mock)
        yield

    def test_telemetry_exception_silenced(self, caplog) -> None:
        with caplog.at_level(logging.DEBUG, logger="huggingface_hub.utils._telemetry"):
            send_telemetry(topic="topic")
            self.queue.join()

        records = [r for r in caplog.records if r.name.startswith("huggingface_hub")]

        # Assert debug message with traceback for debug purposes
        assert len(records) == 1
        assert records[0].levelname == "DEBUG"
        assert records[0].name == "huggingface_hub.utils._telemetry"
        assert records[0].getMessage() == "Error while sending telemetry: whatever"



import os
import time
from io import BytesIO

import pytest

from huggingface_hub import HfApi, hf_hub_download, snapshot_download
from huggingface_hub.errors import EntryNotFoundError
from huggingface_hub.utils import SoftTemporaryDirectory, logging

from .testing_utils import repo_name


logger = logging.get_logger(__name__)
MODEL_IDENTIFIER = "hf-internal-testing/hfh-cache-layout"


def get_file_contents(path):
    with open(path) as f:
        content = f.read()

    return content


@pytest.mark.production
class TestCacheFileLayoutHfHubDownload:
    @pytest.mark.skipif(os.name == "nt", reason="Symlinks are deactivated in Windows tests.")
    def test_file_downloaded_in_cache(self):
        for revision, expected_reference in (
            (None, "main"),
            ("file-2", "file-2"),
        ):
            with SoftTemporaryDirectory() as cache:
                hf_hub_download(
                    MODEL_IDENTIFIER,
                    "file_0.txt",
                    cache_dir=cache,
                    revision=revision,
                )

                expected_directory_name = f"models--{MODEL_IDENTIFIER.replace('/', '--')}"
                expected_path = os.path.join(cache, expected_directory_name)

                refs = os.listdir(os.path.join(expected_path, "refs"))
                snapshots = os.listdir(os.path.join(expected_path, "snapshots"))

                # Only reference should be the expected one.
                assert refs == [expected_reference]

                with open(os.path.join(expected_path, "refs", expected_reference)) as f:
                    snapshot_name = f.readline().strip()

                # The `main` reference should point to the only snapshot we have downloaded
                assert snapshots == [snapshot_name]

                snapshot_path = os.path.join(expected_path, "snapshots", snapshot_name)
                snapshot_content = os.listdir(snapshot_path)

                # Only a single file in the snapshot
                assert len(snapshot_content) == 1

                snapshot_content_path = os.path.join(snapshot_path, snapshot_content[0])

                # The snapshot content should link to a blob
                assert os.path.islink(snapshot_content_path)

                resolved_blob_relative = os.readlink(snapshot_content_path)
                resolved_blob_absolute = os.path.normpath(os.path.join(snapshot_path, resolved_blob_relative))

                with open(resolved_blob_absolute) as f:
                    blob_contents = f.readline().strip()

                # The contents of the file should be 'File 0'.
                assert blob_contents == "File 0"

    def test_no_exist_file_is_cached(self):
        revisions = [None, "file-2"]
        expected_references = ["main", "file-2"]
        for revision, expected_reference in zip(revisions, expected_references):
            with SoftTemporaryDirectory() as cache:
                filename = "this_does_not_exist.txt"
                with pytest.raises(EntryNotFoundError):
                    # The file does not exist, so we get an exception.
                    hf_hub_download(MODEL_IDENTIFIER, filename, cache_dir=cache, revision=revision)

                expected_directory_name = f"models--{MODEL_IDENTIFIER.replace('/', '--')}"
                expected_path = os.path.join(cache, expected_directory_name)

                refs = os.listdir(os.path.join(expected_path, "refs"))
                no_exist_snapshots = os.listdir(os.path.join(expected_path, ".no_exist"))

                # Only reference should be `main`.
                assert refs == [expected_reference]

                with open(os.path.join(expected_path, "refs", expected_reference)) as f:
                    snapshot_name = f.readline().strip()

                # The `main` reference should point to the only snapshot we have downloaded
                assert no_exist_snapshots == [snapshot_name]

                no_exist_path = os.path.join(expected_path, ".no_exist", snapshot_name)
                no_exist_content = os.listdir(no_exist_path)

                # Only a single file in the no_exist snapshot
                assert len(no_exist_content) == 1

                # The no_exist content should be our file
                assert no_exist_content[0] == filename

                with open(os.path.join(no_exist_path, filename)) as f:
                    content = f.read().strip()

                # The contents of the file should be empty.
                assert content == ""

    def test_file_download_happens_once(self):
        # Tests that a file is only downloaded once if it's not updated.
        with SoftTemporaryDirectory() as cache:
            path = hf_hub_download(MODEL_IDENTIFIER, "file_0.txt", cache_dir=cache)
            creation_time_0 = os.path.getmtime(path)

            time.sleep(2)

            path = hf_hub_download(MODEL_IDENTIFIER, "file_0.txt", cache_dir=cache)
            creation_time_1 = os.path.getmtime(path)

            assert creation_time_0 == creation_time_1

    @pytest.mark.skipif(os.name == "nt", reason="Symlinks are deactivated in Windows tests.")
    def test_file_download_happens_once_intra_revision(self):
        # Tests that a file is only downloaded once if it's not updated, even across different revisions.

        with SoftTemporaryDirectory() as cache:
            path = hf_hub_download(MODEL_IDENTIFIER, "file_0.txt", cache_dir=cache)
            creation_time_0 = os.path.getmtime(path)

            time.sleep(2)

            path = hf_hub_download(MODEL_IDENTIFIER, "file_0.txt", cache_dir=cache, revision="file-2")
            creation_time_1 = os.path.getmtime(path)

            assert creation_time_0 == creation_time_1

    @pytest.mark.skipif(os.name == "nt", reason="Symlinks are deactivated in Windows tests.")
    def test_multiple_refs_for_same_file(self):
        with SoftTemporaryDirectory() as cache:
            hf_hub_download(MODEL_IDENTIFIER, "file_0.txt", cache_dir=cache)
            hf_hub_download(MODEL_IDENTIFIER, "file_0.txt", cache_dir=cache, revision="file-2")

            expected_directory_name = f"models--{MODEL_IDENTIFIER.replace('/', '--')}"
            expected_path = os.path.join(cache, expected_directory_name)

            refs = os.listdir(os.path.join(expected_path, "refs"))
            refs.sort()

            snapshots = os.listdir(os.path.join(expected_path, "snapshots"))
            snapshots.sort()

            # Directory should contain two revisions
            assert refs == ["file-2", "main"]

            refs_contents = [get_file_contents(os.path.join(expected_path, "refs", f)) for f in refs]
            refs_contents.sort()

            # snapshots directory should contain two snapshots
            assert refs_contents == snapshots

            snapshot_links = [
                os.readlink(os.path.join(expected_path, "snapshots", filename, "file_0.txt")) for filename in snapshots
            ]

            # All snapshot links should point to the same file.
            assert snapshot_links[0] == snapshot_links[1]


@pytest.mark.production
class TestCacheFileLayoutSnapshotDownload:
    @pytest.mark.skipif(os.name == "nt", reason="Symlinks are deactivated in Windows tests.")
    def test_file_downloaded_in_cache(self):
        with SoftTemporaryDirectory() as cache:
            snapshot_download(MODEL_IDENTIFIER, cache_dir=cache)

            expected_directory_name = f"models--{MODEL_IDENTIFIER.replace('/', '--')}"
            expected_path = os.path.join(cache, expected_directory_name)

            refs = os.listdir(os.path.join(expected_path, "refs"))

            snapshots = os.listdir(os.path.join(expected_path, "snapshots"))
            snapshots.sort()

            # Directory should contain two revisions
            assert refs == ["main"]

            ref_content = get_file_contents(os.path.join(expected_path, "refs", refs[0]))

            # snapshots directory should contain two snapshots
            assert [ref_content] == snapshots

            snapshot_path = os.path.join(expected_path, "snapshots", snapshots[0])

            files_in_snapshot = os.listdir(snapshot_path)

            snapshot_links = [os.readlink(os.path.join(snapshot_path, filename)) for filename in files_in_snapshot]

            resolved_snapshot_links = [os.path.normpath(os.path.join(snapshot_path, link)) for link in snapshot_links]

            assert all([os.path.isfile(link) for link in resolved_snapshot_links])

    @pytest.mark.skipif(os.name == "nt", reason="Symlinks are deactivated in Windows tests.")
    def test_file_downloaded_in_cache_several_revisions(self):
        with SoftTemporaryDirectory() as cache:
            snapshot_download(MODEL_IDENTIFIER, cache_dir=cache, revision="file-3")
            snapshot_download(MODEL_IDENTIFIER, cache_dir=cache, revision="file-2")

            expected_directory_name = f"models--{MODEL_IDENTIFIER.replace('/', '--')}"
            expected_path = os.path.join(cache, expected_directory_name)

            refs = os.listdir(os.path.join(expected_path, "refs"))
            refs.sort()

            snapshots = os.listdir(os.path.join(expected_path, "snapshots"))
            snapshots.sort()

            # Directory should contain two revisions
            assert refs == ["file-2", "file-3"]

            refs_content = [get_file_contents(os.path.join(expected_path, "refs", ref)) for ref in refs]
            refs_content.sort()

            # snapshots directory should contain two snapshots
            assert refs_content == snapshots

            snapshots_paths = [os.path.join(expected_path, "snapshots", s) for s in snapshots]

            files_in_snapshots = {s: os.listdir(s) for s in snapshots_paths}
            links_in_snapshots = {
                k: [os.readlink(os.path.join(k, _v)) for _v in v] for k, v in files_in_snapshots.items()
            }

            resolved_snapshots_links = {
                k: [os.path.normpath(os.path.join(k, link)) for link in v] for k, v in links_in_snapshots.items()
            }

            all_links = [b for a in resolved_snapshots_links.values() for b in a]
            all_unique_links = set(all_links)

            # [ 100]  .
            # ├── [ 140]  blobs
            # │   ├── [   7]  4475433e279a71203927cbe80125208a3b5db560
            # │   ├── [   7]  50fcd26d6ce3000f9d5f12904e80eccdc5685dd1
            # │   ├── [   7]  80146afc836c60e70ba67933fec439ab05b478f6
            # │   ├── [   7]  8cf9e18f080becb674b31c21642538269fe886a4
            # │   └── [1.1K]  ac481c8eb05e4d2496fbe076a38a7b4835dd733d
            # ├── [  80]  refs
            # │   ├── [  40]  file-2
            # │   └── [  40]  file-3
            # └── [  80]  snapshots
            #     ├── [ 120]  5e23cb3ae7f904919a442e1b27dcddae6c6bc292
            #     │   ├── [  52]  file_0.txt -> ../../blobs/80146afc836c60e70ba67933fec439ab05b478f6
            #     │   ├── [  52]  file_1.txt -> ../../blobs/50fcd26d6ce3000f9d5f12904e80eccdc5685dd1
            #     │   ├── [  52]  file_2.txt -> ../../blobs/4475433e279a71203927cbe80125208a3b5db560
            #     │   └── [  52]  .gitattributes -> ../../blobs/ac481c8eb05e4d2496fbe076a38a7b4835dd733d
            #     └── [ 120]  78aa2ebdb60bba086496a8792ba506e58e587b4c
            #         ├── [  52]  file_0.txt -> ../../blobs/80146afc836c60e70ba67933fec439ab05b478f6
            #         ├── [  52]  file_1.txt -> ../../blobs/50fcd26d6ce3000f9d5f12904e80eccdc5685dd1
            #         ├── [  52]  file_3.txt -> ../../blobs/8cf9e18f080becb674b31c21642538269fe886a4
            #         └── [  52]  .gitattributes -> ../../blobs/ac481c8eb05e4d2496fbe076a38a7b4835dd733d

            # Across the two revisions, there should be 8 total links
            assert len(all_links) == 8

            # Across the two revisions, there should only be 5 unique files.
            assert len(all_unique_links) == 5


class TestReferenceUpdates:
    def test_update_reference(self, api: HfApi):
        repo_id = api.create_repo(repo_name(), exist_ok=True).repo_id

        try:
            api.upload_file(path_or_fileobj=BytesIO(b"Some string"), path_in_repo="file.txt", repo_id=repo_id)

            with SoftTemporaryDirectory() as cache:
                hf_hub_download(repo_id, "file.txt", cache_dir=cache)

                expected_directory_name = f"models--{repo_id.replace('/', '--')}"
                expected_path = os.path.join(cache, expected_directory_name)

                refs = os.listdir(os.path.join(expected_path, "refs"))

                # Directory should contain two revisions
                assert refs == ["main"]

                initial_ref_content = get_file_contents(os.path.join(expected_path, "refs", refs[0]))

                # Upload a new file on the same branch
                api.upload_file(
                    path_or_fileobj=BytesIO(b"Some new string"),
                    path_in_repo="file.txt",
                    repo_id=repo_id,
                )

                hf_hub_download(repo_id, "file.txt", cache_dir=cache)

                final_ref_content = get_file_contents(os.path.join(expected_path, "refs", refs[0]))

                # The `main` reference should point to two different, but existing snapshots which contain
                # a 'file.txt'
                assert initial_ref_content != final_ref_content
                assert os.path.isdir(os.path.join(expected_path, "snapshots", initial_ref_content))
                assert os.path.isfile(os.path.join(expected_path, "snapshots", initial_ref_content, "file.txt"))
                assert os.path.isdir(os.path.join(expected_path, "snapshots", final_ref_content))
                assert os.path.isfile(os.path.join(expected_path, "snapshots", final_ref_content, "file.txt"))
        except Exception:
            raise
        finally:
            api.delete_repo(repo_id)



import os
from contextlib import ExitStack
from typing import Generator, Protocol

import pytest
from _pytest.fixtures import SubRequest
from _pytest.skipping import evaluate_skip_marks

import huggingface_hub
from huggingface_hub import HfApi, RepoUrl, constants
from huggingface_hub.utils import SoftTemporaryDirectory, _detect_agent, logging
from huggingface_hub.utils._runtime import is_package_available

from .testing_constants import (
    ENDPOINT_PRODUCTION,
    ENDPOINT_PRODUCTION_URL_SCHEME,
    ENDPOINT_STAGING,
    TOKEN,
)
from .testing_utils import repo_name


@pytest.fixture(autouse=True, scope="function")
def patch_constants(mocker):
    with SoftTemporaryDirectory() as cache_dir:
        mocker.patch.object(constants, "HF_HOME", cache_dir)
        mocker.patch.object(constants, "HF_HUB_CACHE", os.path.join(cache_dir, "hub"))
        mocker.patch.object(constants, "HF_XET_CACHE", os.path.join(cache_dir, "xet"))
        mocker.patch.object(constants, "HUGGINGFACE_HUB_CACHE", os.path.join(cache_dir, "hub"))
        mocker.patch.object(constants, "HF_ASSETS_CACHE", os.path.join(cache_dir, "assets"))
        mocker.patch.object(constants, "HF_TOKEN_PATH", os.path.join(cache_dir, "token"))
        mocker.patch.object(constants, "HF_STORED_TOKENS_PATH", os.path.join(cache_dir, "stored_tokens"))
        yield


@pytest.fixture(autouse=True)
def reset_oauth_refresh_state():
    """`get_token()` caches its OAuth refresh decision in process-global state: reset between tests."""
    from huggingface_hub.utils import _auth

    _auth._OAUTH_REFRESH_CACHE = None
    _auth._OAUTH_REFRESH_WARNED = False
    yield


@pytest.fixture(autouse=True)
def xet_mode(request: SubRequest, monkeypatch: pytest.MonkeyPatch) -> None:
    """Make Xet usage explicit and deterministic, locally and in CI.

    Three modes:
    - `@pytest.mark.xet`: test requires `hf_xet` => skipped when it is not installed,
      Xet force-enabled otherwise.
    - `@pytest.mark.no_xet`: test must run without `hf_xet` (e.g. legacy LFS behavior)
      => skipped when it is installed.
    - unmarked: test must work regardless of Xet => nothing is forced; the test runs
      with whatever the environment provides. CI runs unmarked tests both with and
      without `hf_xet` installed.
    """
    xet = request.node.get_closest_marker("xet") is not None
    no_xet = request.node.get_closest_marker("no_xet") is not None
    if xet and no_xet:
        pytest.fail("A test cannot be marked with both `xet` and `no_xet`.")
    if xet:
        if not is_package_available("hf_xet"):
            pytest.skip("Test requires `hf_xet` (marked with `pytest.mark.xet`)")
        monkeypatch.setattr(constants, "HF_HUB_DISABLE_XET", False)
        monkeypatch.delenv("HF_HUB_DISABLE_XET", raising=False)
    elif no_xet:
        if is_package_available("hf_xet"):
            pytest.skip("Test must run without `hf_xet` installed (marked with `pytest.mark.no_xet`)")


@pytest.fixture(autouse=True)
def _clean_cli_env(monkeypatch: pytest.MonkeyPatch) -> None:
    """Deterministic baseline: agent detection disabled, no ANSI colors, reset output mode, fixed terminal width."""
    # Pin an empty agent registry so detection never hits the network and always reports "no agent",
    # regardless of any agent env var that may be set on the host running the tests.
    monkeypatch.setattr(_detect_agent, "_registry", _detect_agent._EMPTY_REGISTRY)
    monkeypatch.setenv("NO_COLOR", "1")
    # Pin terminal width so adaptive truncation produces deterministic output
    # regardless of the runner's actual terminal size. `shutil.get_terminal_size`
    # honors `$COLUMNS` first.
    monkeypatch.setenv("COLUMNS", "200")
    from huggingface_hub.cli._output import out

    out.set_mode()
    out.set_no_truncate(False)


logger = logging.get_logger(__name__)


@pytest.fixture(autouse=True)
def disable_symlinks_on_windows_ci(monkeypatch: pytest.MonkeyPatch) -> None:
    class FakeSymlinkDict(dict):
        def __contains__(self, __o: object) -> bool:
            return True  # consider any `cache_dir` to be already checked

        def __getitem__(self, __key: str) -> bool:
            return False  # symlinks are never supported

    if os.name == "nt" and os.environ.get("DISABLE_SYMLINKS_IN_WINDOWS_TESTS"):
        monkeypatch.setattr(
            huggingface_hub.file_download,
            "_are_symlinks_supported_in_dir",
            FakeSymlinkDict(),
        )


@pytest.fixture(autouse=True)
def disable_experimental_warnings(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(huggingface_hub.constants, "HF_HUB_DISABLE_EXPERIMENTAL_WARNING", True)


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": ["authorization", "user-agent", "cookie"],
        "ignore_localhost": True,
        "path_transformer": lambda path: path + ".yaml",
    }


@pytest.fixture(autouse=True)
def clear_lru_cache():
    from huggingface_hub.inference._providers.hf_inference import _check_supported_task

    _check_supported_task.cache_clear()
    yield
    _check_supported_task.cache_clear()


@pytest.fixture(autouse=True)
def production_endpoint_marker(request: SubRequest, monkeypatch: pytest.MonkeyPatch) -> None:
    """Point the Hub at production for tests marked with `@pytest.mark.production`.

    Mirrors the former `with_production_testing` decorator: patches both `ENDPOINT` and the
    download URL template so that bare `HfApi()` instances (and downloads) target production.
    The `api_prod` fixture passes the production endpoint explicitly and does not rely on this.
    """
    if request.node.get_closest_marker("production") is None:
        return
    monkeypatch.setattr(constants, "ENDPOINT", ENDPOINT_PRODUCTION)
    monkeypatch.setattr(constants, "HUGGINGFACE_CO_URL_TEMPLATE", ENDPOINT_PRODUCTION_URL_SCHEME)


@pytest.fixture(autouse=True)
def expect_deprecation_marker(request: SubRequest) -> Generator[None, None, None]:
    """Assert that a test marked `@pytest.mark.deprecated(...)` emits the expected `FutureWarning`s.

    Each argument is a function name; the test must emit a `FutureWarning` mentioning it (the suite runs
    with `-Werror::FutureWarning`, so an unraised warning fails the test). Example:
    ```py
    @pytest.mark.deprecated("duplicate_space", "duplicate_repo")
    def test_duplicate_space(...): ...
    ```
    """
    function_names = [name for marker in request.node.iter_markers("deprecated") for name in marker.args]
    # If the test is statically skipped (`@pytest.mark.skip`/`skipif`), its body never runs and no
    # warning is emitted; entering `pytest.warns` would then fail the item at teardown instead of
    # reporting it as skipped. `evaluate_skip_marks` returns a truthy result only when it will skip.
    if not function_names or evaluate_skip_marks(request.node):
        yield
        return
    with ExitStack() as stack:
        for function_name in function_names:
            stack.enter_context(pytest.warns(FutureWarning, match=rf".*'{function_name}'.*"))
        yield


@pytest.fixture(autouse=True)
def git_lfs_marker(request: SubRequest) -> None:
    """Skip tests marked `@pytest.mark.git_lfs` unless `RUN_GIT_LFS_TESTS` is truthy (git-lfs needed)."""
    if request.node.get_closest_marker("git_lfs") is not None and os.environ.get("RUN_GIT_LFS_TESTS") != "1":
        pytest.skip("git-lfs test (set RUN_GIT_LFS_TESTS=1 to run)")


@pytest.fixture(scope="session")
def api() -> HfApi:
    """A staging `HfApi` client authenticated with the CI token."""
    return HfApi(endpoint=ENDPOINT_STAGING, token=TOKEN)


@pytest.fixture(scope="session")
def api_prod() -> HfApi:
    """An unauthenticated `HfApi` client targeting the production Hub (read-only tests)."""
    return HfApi(endpoint=ENDPOINT_PRODUCTION)


class RepoFactory(Protocol):
    """Type of the `repo_factory` fixture: create a temporary repo on staging and return its `RepoUrl`.

    The created repo is automatically deleted when the test ends.
    """

    def __call__(self, repo_type: str = "model", **kwargs) -> RepoUrl: ...


@pytest.fixture
def repo_factory(api: HfApi) -> Generator[RepoFactory, None, None]:
    """Create temporary repos on staging and delete them automatically when the test ends.

    Example:
    ```py
    def test_something(api: HfApi, repo_factory):
        repo_url = repo_factory()              # a model repo
        dataset_url = repo_factory("dataset")  # a dataset repo
    ```
    """
    created: list[tuple[str, str]] = []

    def _factory(repo_type: str = "model", **kwargs) -> RepoUrl:
        if repo_type == "space" and "space_sdk" not in kwargs:
            kwargs["space_sdk"] = "gradio"
        repo_url = api.create_repo(repo_id=repo_name(prefix=repo_type), repo_type=repo_type, **kwargs)
        created.append((repo_url.repo_id, repo_type))
        return repo_url

    yield _factory

    for repo_id, repo_type in created:
        try:
            api.delete_repo(repo_id=repo_id, repo_type=repo_type)
        except Exception:
            pass



# Copyright 2026 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Device Code OAuth (RFC 8628) for browser-based login, plus access token refresh.

The flow: the client requests a device code, displays a URL and a short user code, the user
authorizes in a browser, and the client polls ``POST {ENDPOINT}/oauth/token`` until a token is
issued. Access tokens may come with a refresh token, used to renew them transparently (see
``utils/_auth.py::get_token``).

This module is self-contained protocol logic: no printing, no persistence. Interactive flows
live in ``_login.py`` (human/library) and ``cli/auth.py`` (machine-readable event stream).
"""

import time
from collections.abc import Callable
from typing import TypedDict, cast

import httpx

from .. import constants
from ..errors import DeviceCodeError, OAuthErrorCode
from ._http import get_session, hf_raise_for_status


_DEVICE_CODE_GRANT_TYPE = "urn:ietf:params:oauth:grant-type:device_code"
_REFRESH_TOKEN_GRANT_TYPE = "refresh_token"


class DeviceCodeInfo(TypedDict):
    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str  # falls back to verification_uri if the server omits it
    interval: int
    expires_in: int


class OAuthTokenResponse(TypedDict, total=False):
    access_token: str
    refresh_token: str
    expires_in: int
    token_type: str


def request_device_code() -> DeviceCodeInfo:
    """Request a device code from the Hub's OAuth device authorization endpoint.

    The returned dict is normalized: `interval`, `expires_in` and `verification_uri_complete`
    are always set (server values, or sensible defaults).

    Raises:
        [`DeviceCodeError`]: If the request fails.
    """
    try:
        response = get_session().post(
            f"{constants.ENDPOINT}/oauth/device",
            data={"client_id": constants.DEVICE_CODE_OAUTH_CLIENT_ID},
            timeout=constants.HF_HUB_DOWNLOAD_TIMEOUT,
        )
        hf_raise_for_status(response)
    except httpx.HTTPError as e:
        raise DeviceCodeError(f"Failed to request device code from {constants.ENDPOINT}/oauth/device: {e}") from e
    info = response.json()
    # `interval` is optional per RFC 8628 (5s is the spec-mandated fallback); `expires_in` is
    # required but defaulted defensively so polling stays bounded if a server omits it.
    info.setdefault("interval", 5)
    info.setdefault("expires_in", 900)
    if not info.get("verification_uri_complete"):
        info["verification_uri_complete"] = info["verification_uri"]
    return cast(DeviceCodeInfo, info)


def poll_device_token(
    device_info: DeviceCodeInfo, *, on_pending: Callable[[], None] | None = None
) -> OAuthTokenResponse:
    """Poll the token endpoint until the user authorizes the device.

    Args:
        device_info (`DeviceCodeInfo`):
            The device authorization response from [`request_device_code`].
        on_pending (`Callable`, *optional*):
            Called after each "authorization pending" response (e.g. to print a progress dot).

    Returns:
        `OAuthTokenResponse`: the full token response: `access_token`, and optionally
        `refresh_token` and `expires_in`.

    Raises:
        [`DeviceCodeError`]: If authorization is denied, the device code expires, or the server
            returns an unexpected OAuth error.
    """
    interval = device_info["interval"]
    deadline = time.monotonic() + device_info["expires_in"]
    while time.monotonic() < deadline:
        # Inconclusive responses (network blip, 5xx, gateway error page, rate limiting) must not
        # abort the login: keep polling until the device code expires (RFC 8628 section 3.5).
        # The deadline bounds the total wait even if the endpoint is genuinely broken.
        data = None
        try:
            response = get_session().post(
                f"{constants.ENDPOINT}/oauth/token",
                data={
                    "grant_type": _DEVICE_CODE_GRANT_TYPE,
                    "device_code": device_info["device_code"],
                    "client_id": constants.DEVICE_CODE_OAUTH_CLIENT_ID,
                },
                timeout=constants.HF_HUB_DOWNLOAD_TIMEOUT,
            )
            if response.status_code < 500:
                data = response.json()
        except (httpx.HTTPError, ValueError):
            pass

        if data is not None:
            if "access_token" in data:
                return cast(OAuthTokenResponse, data)

            match data.get("error"):
                case None:
                    pass  # JSON without an OAuth `error` field (proxy error page, ...): transient
                case OAuthErrorCode.AUTHORIZATION_PENDING:
                    if on_pending is not None:
                        on_pending()
                case OAuthErrorCode.SLOW_DOWN:
                    interval += 5
                case OAuthErrorCode.EXPIRED_TOKEN:
                    raise DeviceCodeError(
                        "Device code expired. Please try again.", error_code=OAuthErrorCode.EXPIRED_TOKEN
                    )
                case OAuthErrorCode.ACCESS_DENIED:
                    raise DeviceCodeError(
                        "Authorization was denied. Please try again.", error_code=OAuthErrorCode.ACCESS_DENIED
                    )
                case error:
                    raise DeviceCodeError(
                        f"OAuth error: {error} - {data.get('error_description', '')}", error_code=error
                    )

        time.sleep(interval)

    raise DeviceCodeError("Device code expired (timeout). Please try again.", error_code=OAuthErrorCode.EXPIRED_TOKEN)


def refresh_access_token(refresh_token: str) -> OAuthTokenResponse:
    """Exchange a refresh token for a new access token.

    Returns:
        `OAuthTokenResponse`: the full token response: `access_token`, and optionally a rotated
        `refresh_token` and `expires_in`.

    Raises:
        [`DeviceCodeError`]: If the server rejects the refresh (`error_code="invalid_grant"` when
            the refresh token is expired or revoked) or returns an unexpected response.
    """
    try:
        response = get_session().post(
            f"{constants.ENDPOINT}/oauth/token",
            data={
                "grant_type": _REFRESH_TOKEN_GRANT_TYPE,
                "refresh_token": refresh_token,
                "client_id": constants.DEVICE_CODE_OAUTH_CLIENT_ID,
            },
            # An explicit timeout is critical here: this runs inside `get_token()`, so a hung
            # request would otherwise block every Hub call in the process.
            timeout=constants.HF_HUB_DOWNLOAD_TIMEOUT,
        )
    except httpx.HTTPError as e:
        raise DeviceCodeError(f"Failed to refresh access token: {e}") from e
    data = _parse_token_response(response)
    if "access_token" in data:
        return cast(OAuthTokenResponse, data)
    error = data.get("error")
    raise DeviceCodeError(
        f"Failed to refresh access token: {error or response.status_code} - {data.get('error_description', '')}",
        error_code=error,
    )


def _parse_token_response(response: httpx.Response) -> dict:
    try:
        return response.json()
    except ValueError as e:
        raise DeviceCodeError(
            f"Failed to parse response from {constants.ENDPOINT}/oauth/token "
            f"(status {response.status_code}): {response.text[:500]}"
        ) from e



import json
from functools import lru_cache
from pathlib import Path
from typing import Any
from urllib.parse import urlparse, urlunparse

from huggingface_hub import constants
from huggingface_hub.hf_api import InferenceProviderMapping
from huggingface_hub.inference._common import (
    MimeBytes,
    RequestParameters,
    _b64_encode,
    _bytes_to_dict,
    _open_as_mime_bytes,
)
from huggingface_hub.inference._providers._common import TaskProviderHelper, filter_none
from huggingface_hub.utils import build_hf_headers, get_session, get_token, hf_raise_for_status


class HFInferenceTask(TaskProviderHelper):
    """Base class for HF Inference API tasks."""

    def __init__(self, task: str):
        super().__init__(
            provider="hf-inference",
            base_url=constants.INFERENCE_PROXY_TEMPLATE.format(provider="hf-inference"),
            task=task,
        )

    def _prepare_api_key(self, api_key: str | None) -> str:
        # special case: for HF Inference we allow not providing an API key
        return api_key or get_token()  # type: ignore

    def _prepare_mapping_info(self, model: str | None) -> InferenceProviderMapping:
        if model is not None and model.startswith(("http://", "https://")):
            return InferenceProviderMapping(
                provider="hf-inference", providerId=model, hf_model_id=model, task=self.task, status="live"
            )
        model_id = model if model is not None else _fetch_recommended_models().get(self.task)
        if model_id is None:
            raise ValueError(
                f"Task {self.task} has no recommended model for HF Inference. Please specify a model"
                " explicitly. Visit https://huggingface.co/tasks for more info."
            )
        _check_supported_task(model_id, self.task)
        return InferenceProviderMapping(
            provider="hf-inference", providerId=model_id, hf_model_id=model_id, task=self.task, status="live"
        )

    def _prepare_url(self, api_key: str, mapped_model: str) -> str:
        # hf-inference provider can handle URLs (e.g. Inference Endpoints or TGI deployment)
        if mapped_model.startswith(("http://", "https://")):
            return mapped_model
        return (
            # Feature-extraction and sentence-similarity are the only cases where we handle models with several tasks.
            f"{self.base_url}/models/{mapped_model}/pipeline/{self.task}"
            if self.task in ("feature-extraction", "sentence-similarity")
            # Otherwise, we use the default endpoint
            else f"{self.base_url}/models/{mapped_model}"
        )

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        if isinstance(inputs, bytes):
            raise ValueError(f"Unexpected binary input for task {self.task}.")
        if isinstance(inputs, Path):
            raise ValueError(f"Unexpected path input for task {self.task} (got {inputs})")
        return filter_none({"inputs": inputs, "parameters": parameters})


class HFInferenceBinaryInputTask(HFInferenceTask):
    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        return None

    def _prepare_payload_as_bytes(
        self,
        inputs: Any,
        parameters: dict,
        provider_mapping_info: InferenceProviderMapping,
        extra_payload: dict | None,
    ) -> MimeBytes | None:
        parameters = filter_none(parameters)
        extra_payload = extra_payload or {}
        has_parameters = len(parameters) > 0 or len(extra_payload) > 0

        # Raise if not a binary object or a local path or a URL.
        if not isinstance(inputs, (bytes, Path)) and not isinstance(inputs, str):
            raise ValueError(f"Expected binary inputs or a local path or a URL. Got {inputs}")

        # Send inputs as raw content when no parameters are provided
        if not has_parameters:
            return _open_as_mime_bytes(inputs)

        # Otherwise encode as b64
        return MimeBytes(
            json.dumps({"inputs": _b64_encode(inputs), "parameters": parameters, **extra_payload}).encode("utf-8"),
            mime_type="application/json",
        )


class HFInferenceConversational(HFInferenceTask):
    def __init__(self):
        super().__init__("conversational")

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        payload = filter_none(parameters)
        mapped_model = provider_mapping_info.provider_id
        payload_model = parameters.get("model") or mapped_model

        if payload_model is None or payload_model.startswith(("http://", "https://")):
            payload_model = "dummy"

        response_format = parameters.get("response_format")
        if isinstance(response_format, dict) and response_format.get("type") == "json_schema":
            payload["response_format"] = {
                "type": "json_object",
                "value": response_format["json_schema"]["schema"],
            }
        return {**payload, "model": payload_model, "messages": inputs}

    def _prepare_url(self, api_key: str, mapped_model: str) -> str:
        base_url = (
            mapped_model
            if mapped_model.startswith(("http://", "https://"))
            else f"{constants.INFERENCE_PROXY_TEMPLATE.format(provider='hf-inference')}/models/{mapped_model}"
        )
        return _build_chat_completion_url(base_url)


def _build_chat_completion_url(model_url: str) -> str:
    parsed = urlparse(model_url)
    path = parsed.path.rstrip("/")

    # If the path already ends with /chat/completions, we're done!
    if path.endswith("/chat/completions"):
        return model_url

    # Append /chat/completions if not already present
    if path.endswith("/v1"):
        new_path = path + "/chat/completions"
    # If path was empty or just "/", set the full path
    elif not path:
        new_path = "/v1/chat/completions"
    # Append /v1/chat/completions if not already present
    else:
        new_path = path + "/v1/chat/completions"

    # Reconstruct the URL with the new path and original query parameters.
    new_parsed = parsed._replace(path=new_path)
    return str(urlunparse(new_parsed))


@lru_cache(maxsize=1)
def _fetch_recommended_models() -> dict[str, str | None]:
    response = get_session().get(f"{constants.ENDPOINT}/api/tasks", headers=build_hf_headers())
    hf_raise_for_status(response)
    return {task: next(iter(details["widgetModels"]), None) for task, details in response.json().items()}


@lru_cache(maxsize=None)
def _check_supported_task(model: str, task: str) -> None:
    from huggingface_hub.hf_api import HfApi

    model_info = HfApi().model_info(model)
    pipeline_tag = model_info.pipeline_tag
    tags = model_info.tags or []
    is_conversational = "conversational" in tags
    if task in ("text-generation", "conversational"):
        if pipeline_tag == "text-generation":
            # text-generation + conversational tag -> both tasks allowed
            if is_conversational:
                return
            # text-generation without conversational tag -> only text-generation allowed
            if task == "text-generation":
                return
            raise ValueError(f"Model '{model}' doesn't support task '{task}'.")

    if pipeline_tag == "text2text-generation":
        if task == "text-generation":
            return
        raise ValueError(f"Model '{model}' doesn't support task '{task}'.")

    if pipeline_tag == "image-text-to-text":
        if is_conversational and task == "conversational":
            return  # Only conversational allowed if tagged as conversational
        raise ValueError("Non-conversational image-text-to-text task is not supported.")

    if (
        task in ("feature-extraction", "sentence-similarity")
        and pipeline_tag in ("feature-extraction", "sentence-similarity")
        and task in tags
    ):
        # feature-extraction and sentence-similarity are interchangeable for HF Inference
        return

    # For all other tasks, just check pipeline tag
    if pipeline_tag != task:
        raise ValueError(
            f"Model '{model}' doesn't support task '{task}'. Supported tasks: '{pipeline_tag}', got: '{task}'"
        )
    return


class HFInferenceFeatureExtractionTask(HFInferenceTask):
    def __init__(self):
        super().__init__("feature-extraction")

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        if isinstance(inputs, bytes):
            raise ValueError(f"Unexpected binary input for task {self.task}.")
        if isinstance(inputs, Path):
            raise ValueError(f"Unexpected path input for task {self.task} (got {inputs})")

        # Parameters are sent at root-level for feature-extraction task
        # See specs: https://github.com/huggingface/huggingface.js/blob/main/packages/tasks/src/tasks/feature-extraction/spec/input.json
        return {"inputs": inputs, **filter_none(parameters)}

    def get_response(self, response: bytes | dict, request_params: RequestParameters | None = None) -> Any:
        if isinstance(response, bytes):
            return _bytes_to_dict(response)
        return response



# Copyright 2026 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the minimal Click framework (`huggingface_hub.cli._framework`).

These exercise the framework in isolation (no `hf` commands), covering the
annotation -> click.Parameter conversion, value handling at invocation time,
help-record rendering, and the group/command decorator API.
"""

import enum
from pathlib import Path
from typing import Annotated, Literal

import click
import pytest
from click.testing import CliRunner

from huggingface_hub.cli._framework import (
    Argument,
    HfGroup,
    Option,
    build_command,
    get_command_name,
)


class Color(str, enum.Enum):
    red = "red"
    blue = "blue"


def _params(func) -> dict:
    return {p.name: p for p in build_command(func, name="x").params}


def test_get_command_name_kebab_cases():
    assert get_command_name("force_download") == "force-download"
    assert get_command_name("type") == "type"


class TestParameterBuilding:
    def test_required_argument(self):
        def f(repo_id: Annotated[str, Argument()]): ...

        param = _params(f)["repo_id"]
        assert isinstance(param, click.Argument) and param.required and param.type is click.STRING

    def test_optional_list_argument_uses_nargs_minus_one(self):
        def f(files: Annotated[list[str] | None, Argument()] = None): ...

        param = _params(f)["files"]
        assert isinstance(param, click.Argument) and not param.required and param.nargs == -1

    def test_bool_option_without_decl_generates_toggle(self):
        def f(force: Annotated[bool, Option()] = False): ...

        param = _params(f)["force"]
        assert param.is_flag and param.opts == ["--force"] and param.secondary_opts == ["--no-force"]

    def test_bool_option_with_explicit_decl_is_single_flag(self):
        def f(meta: Annotated[bool, Option("--meta")] = False): ...

        param = _params(f)["meta"]
        assert param.is_flag and param.opts == ["--meta"] and not param.secondary_opts

    def test_optional_str_option_defaults_none(self):
        def f(revision: Annotated[str | None, Option()] = None): ...

        param = _params(f)["revision"]
        assert isinstance(param, click.Option) and not param.required and param.default is None

    def test_int_option_with_min_is_intrange(self):
        def f(workers: Annotated[int, Option(min=1)] = 1): ...

        param = _params(f)["workers"]
        assert isinstance(param.type, click.IntRange) and param.type.min == 1

    def test_float_option_with_min_is_floatrange(self):
        def f(rate: Annotated[float, Option(min=1)] = 2.0): ...

        param = _params(f)["rate"]
        assert isinstance(param.type, click.FloatRange) and param.type.min == 1

    def test_argument_forwards_is_eager(self):
        def f(repo_id: Annotated[str, Argument(is_eager=True)]): ...

        assert _params(f)["repo_id"].is_eager

    def test_non_str_enum_raises_at_build_time(self):
        class Number(enum.Enum):
            one = 1
            two = 2

        def f(n: Annotated[Number, Option("--n")] = Number.one): ...

        with pytest.raises(TypeError, match="only str-valued enums"):
            build_command(f, name="x")

    def test_enum_option_builds_choice_from_values(self):
        def f(color: Annotated[Color, Option("--color")] = Color.red): ...

        param = _params(f)["color"]
        assert isinstance(param.type, click.Choice) and list(param.type.choices) == ["red", "blue"]

    def test_literal_option_builds_choice(self):
        def f(region: Annotated[Literal["us", "eu"], Option("--region")] = "us"): ...

        param = _params(f)["region"]
        assert isinstance(param.type, click.Choice) and list(param.type.choices) == ["us", "eu"]

    def test_path_option(self):
        def f(path: Annotated[Path | None, Option()] = None): ...

        assert isinstance(_params(f)["path"].type, click.Path)

    def test_multi_decl_option_name_matches_python_param(self):
        # The Python name is prepended as the first decl so the handler kwarg lines up
        # regardless of the user-facing flag names.
        def f(volumes: Annotated[list[str] | None, Option("-v", "--volume")] = None): ...

        param = _params(f)["volumes"]
        assert param.name == "volumes" and "-v" in param.opts and "--volume" in param.opts and param.multiple

    def test_explicit_click_type_overrides_inference(self):
        soft_choice = click.Choice(["a", "b"])

        def f(mode: Annotated[str, Option("--mode", click_type=soft_choice)] = "a"): ...

        assert _params(f)["mode"].type is soft_choice

    def test_unsupported_annotation_raises(self):
        def f(bad: Annotated[dict, Option("--bad")] = None): ...

        with pytest.raises(TypeError):
            build_command(f, name="x")


class TestInvocation:
    def test_enum_flag_and_list_reach_handler(self):
        captured = {}

        def f(
            repo_id: Annotated[str, Argument()],
            color: Annotated[Color, Option("--color")] = Color.red,
            force: Annotated[bool, Option()] = False,
            volumes: Annotated[list[str] | None, Option("-v", "--volume")] = None,
        ):
            captured.update(repo_id=repo_id, color=color, force=force, volumes=volumes)

        result = CliRunner().invoke(
            build_command(f, name="run"), ["r", "--color", "blue", "--force", "-v", "a", "-v", "b"]
        )
        assert result.exit_code == 0, result.output
        assert captured == {"repo_id": "r", "color": Color.blue, "force": True, "volumes": ["a", "b"]}

    def test_enum_member_default_reaches_handler(self):
        class Mode(str, enum.Enum):
            ON = "on"
            AUTO = "auto"

        captured = {}

        def f(mode: Annotated[Mode, Option("--mode")] = Mode.AUTO):
            captured["mode"] = mode

        result = CliRunner().invoke(build_command(f, name="run"), [])
        assert result.exit_code == 0, result.output
        assert captured["mode"] is Mode.AUTO

    def test_value_callback_is_applied(self):
        seen = {}

        def f(name: Annotated[str | None, Option("--name", callback=lambda v: (v or "").upper())] = None):
            seen["name"] = name

        CliRunner().invoke(build_command(f, name="c"), ["--name", "hi"])
        assert seen["name"] == "HI"

    def test_context_is_injected(self):
        seen = {}

        def f(ctx: click.Context, value: Annotated[str | None, Option("--value")] = None):
            seen["is_ctx"] = isinstance(ctx, click.Context)

        CliRunner().invoke(build_command(f, name="c"), ["--value", "1"])
        assert seen.get("is_ctx")


class TestHelpRecords:
    def _help(self, func) -> str:
        return CliRunner().invoke(build_command(func, name="x"), ["--help"]).output

    def test_argument_renders_help_and_required(self):
        def f(repo_id: Annotated[str, Argument(help="the repo")]): ...

        out = self._help(f)
        assert "REPO_ID" in out and "the repo" in out and "[required]" in out

    def test_bool_flag_default_string(self):
        def f(force: Annotated[bool, Option(help="force it")] = False): ...

        out = self._help(f)
        assert "--force / --no-force" in out and "[default: no-force]" in out

    def test_int_option_shows_range_but_argument_does_not(self):
        def f_opt(n: Annotated[int, Option(min=1, help="num")] = 1): ...

        def f_arg(n: Annotated[int, Argument(min=1, help="num")]): ...

        assert "x>=1" in self._help(f_opt)
        assert "x>=1" not in self._help(f_arg)


class TestGroup:
    def test_command_registration_preserves_order_and_aliases(self):
        group = HfGroup(name="root")

        @group.command("list | ls")
        def _list(): ...

        @group.command("info")
        def _info(): ...

        assert "list | ls" in group.commands
        assert group.list_commands(click.Context(group)) == ["list | ls", "info"]

    def test_add_group_sets_name_and_hidden(self):
        root = HfGroup(name="root")
        sub = HfGroup()
        root.add_group(sub, name="repos | repo", hidden=True)
        assert root.commands["repos | repo"] is sub and sub.name == "repos | repo" and sub.hidden

    def test_group_callback_sets_invoke_without_command_and_metavar(self):
        group = HfGroup(name="root")

        @group.group_callback(invoke_without_command=True)
        def _callback(): ...

        assert group.invoke_without_command and group.subcommand_metavar == "[COMMAND] [ARGS]..."


def _fake_typer_marker(kind: str, default=..., *param_decls, **fields):
    """Build an object mimicking typer.models.OptionInfo/ArgumentInfo storage without importing typer.

    Real typer stores the first positional in ``default`` (the flag name in ``Annotated`` usage,
    the default value in old-style usage) and the rest in ``param_decls``.
    """
    cls = type(kind, (), {})
    cls.__module__ = "typer.models"
    marker = cls()
    marker.default = default
    marker.param_decls = param_decls
    for key, value in fields.items():
        setattr(marker, key, value)
    return marker


class TestTyperMarkerShim:
    """Transition shim: `typer_factory` consumers (e.g. transformers) still use typer markers."""

    def test_annotated_option_with_flags(self):
        # Mimics `Annotated[str, typer.Option("--log-level", "-l", help="lvl")] = "info"`.
        def f(
            log_level: Annotated[str, _fake_typer_marker("OptionInfo", "--log-level", "-l", help="lvl")] = "info",
        ): ...

        param = _params(f)["log_level"]
        assert isinstance(param, click.Option)
        assert "--log-level" in param.opts and "-l" in param.opts
        assert param.default == "info" and param.help == "lvl"

    def test_annotated_argument(self):
        def f(model_id: Annotated[str, _fake_typer_marker("ArgumentInfo", help="the id")]): ...

        param = _params(f)["model_id"]
        assert isinstance(param, click.Argument) and param.required and param.help == "the id"

    def test_annotated_bool_builds_toggle(self):
        def f(force: Annotated[bool, _fake_typer_marker("OptionInfo", help="force it")] = False): ...

        param = _params(f)["force"]
        assert param.is_flag and param.opts == ["--force"] and param.secondary_opts == ["--no-force"]

    def test_old_style_marker_as_default(self):
        # Mimics `text: str = typer.Option("hello", "--text")` (marker carries the default value).
        def f(text: str = _fake_typer_marker("OptionInfo", "hello", "--text")): ...

        param = _params(f)["text"]
        assert isinstance(param, click.Option)
        assert param.opts == ["--text"] and not param.required and param.default == "hello"



# Inference code generated from the JSON schema spec in @huggingface/tasks.
#
# See:
#   - script: https://github.com/huggingface/huggingface.js/blob/main/packages/tasks/scripts/inference-codegen.ts
#   - specs:  https://github.com/huggingface/huggingface.js/tree/main/packages/tasks/src/tasks.
from typing import Any

from .base import BaseInferenceType, dataclass_with_extra


@dataclass_with_extra
class ImageTextToVideoTargetSize(BaseInferenceType):
    """The size in pixel of the output video frames."""

    height: int
    width: int


@dataclass_with_extra
class ImageTextToVideoParameters(BaseInferenceType):
    """Additional inference parameters for Image Text To Video"""

    guidance_scale: float | None = None
    """For diffusion models. A higher guidance scale value encourages the model to generate
    videos closely linked to the text prompt at the expense of lower image quality.
    """
    negative_prompt: str | None = None
    """One prompt to guide what NOT to include in video generation."""
    num_frames: float | None = None
    """The num_frames parameter determines how many video frames are generated."""
    num_inference_steps: int | None = None
    """The number of denoising steps. More denoising steps usually lead to a higher quality
    video at the expense of slower inference.
    """
    prompt: str | None = None
    """The text prompt to guide the video generation. Either this or inputs (image) must be
    provided.
    """
    seed: int | None = None
    """Seed for the random number generator."""
    target_size: ImageTextToVideoTargetSize | None = None
    """The size in pixel of the output video frames."""


@dataclass_with_extra
class ImageTextToVideoInput(BaseInferenceType):
    """Inputs for Image Text To Video inference. Either inputs (image) or prompt (in parameters)
    must be provided, or both.
    """

    inputs: str | None = None
    """The input image data as a base64-encoded string. If no `parameters` are provided, you can
    also provide the image data as a raw bytes payload. Either this or prompt must be
    provided.
    """
    parameters: ImageTextToVideoParameters | None = None
    """Additional inference parameters for Image Text To Video"""


@dataclass_with_extra
class ImageTextToVideoOutput(BaseInferenceType):
    """Outputs of inference for the Image Text To Video task"""

    video: Any
    """The generated video returned as raw bytes in the payload."""



# coding=utf-8
# Copyright 2026-present, the HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Check and update SpaceHardware / JobHardware enums against the Hub API.

Fetches the current hardware flavors from the Hub API (list_spaces_hardware,
list_jobs_hardware) and compares them to the hardcoded enums. New flavors are
added; removed flavors are kept for backward compat and marked with ``# legacy``.

Usage:
    python utils/check_hardware_flavors.py           # check only
    python utils/check_hardware_flavors.py --update  # update source files
"""

import argparse
import json
import re
import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, NoReturn

from huggingface_hub.hf_api import HfApi


SPACE_API_PATH = Path(__file__).parents[1] / "src" / "huggingface_hub" / "_space_api.py"
JOBS_API_PATH = Path(__file__).parents[1] / "src" / "huggingface_hub" / "_jobs_api.py"
JOBS_GUIDE_PATH = Path(__file__).parents[1] / "docs" / "source" / "en" / "guides" / "jobs.md"

HARDWARE_TABLE_START = (
    "<!-- HARDWARE_TABLE_START (auto-generated by utils/check_hardware_flavors.py — do not edit) -->"
)
HARDWARE_TABLE_END = "<!-- HARDWARE_TABLE_END -->"


def _make_enum_name(value: str) -> str:
    """Convert a flavor slug like 'cpu-basic' to an enum member name like 'CPU_BASIC'."""
    return value.upper().replace("-", "_")


@dataclass(slots=True)
class _EnumEntry:
    kind: Literal["member", "line"]
    name: str = ""
    value: str = ""
    is_legacy: bool = False
    text: str = ""


def _parse_enum_entries(source: str, class_name: str) -> list[_EnumEntry]:
    """Parse enum body from source, preserving members, section comments, and blank lines."""
    pattern = re.compile(
        rf"class {re.escape(class_name)}\(str, Enum\):(.*?)(?=\nclass |\ndef |\n@dataclass)",
        re.DOTALL,
    )
    match = pattern.search(source)
    if not match:
        print(f"Error: could not find class {class_name} in source.")
        sys.exit(1)

    body = match.group(1)
    # Skip the docstring: find closing """
    doc_end = body.find('"""', body.find('"""') + 3)
    if doc_end == -1:
        print(f"Error: could not find docstring end for class {class_name}.")
        sys.exit(1)
    body_after_doc = body[doc_end + 3 :]

    entries: list[_EnumEntry] = []
    for line in body_after_doc.splitlines():
        m = re.match(r"^\s{4}([A-Z_][A-Z_0-9]*)\s*=\s*\"([^\"]+)\"(.*)$", line)
        if m:
            name, value, rest = m.group(1), m.group(2), m.group(3)
            is_legacy = "# legacy" in rest
            entries.append(_EnumEntry("member", name=name, value=value, is_legacy=is_legacy))
        elif line.strip():
            entries.append(_EnumEntry("line", text=line))
        else:
            entries.append(_EnumEntry("line", text=""))

    # Strip leading blank lines (we always emit one blank line after the docstring).
    while entries and entries[0].kind == "line" and not entries[0].text.strip():
        entries.pop(0)

    return entries


def _build_enum_lines(
    existing_entries: list[_EnumEntry],
    api_values: list[str],
) -> list[str]:
    """Build the list of enum body lines, preserving order, section comments, and blank lines.

    - Existing members keep their position and surrounding comments.
    - Members no longer in the API get ``# legacy`` if not already marked.
    - New API values are appended at the end.
    """
    api_set = set(api_values)
    existing_values = {e.value for e in existing_entries if e.kind == "member"}

    lines: list[str] = []
    for entry in existing_entries:
        if entry.kind == "line":
            lines.append(entry.text)
        else:
            name, value, was_legacy = entry.name, entry.value, entry.is_legacy
            if value not in api_set and not was_legacy:
                lines.append(f'    {name} = "{value}"  # legacy')
            elif was_legacy and value in api_set:
                lines.append(f'    {name} = "{value}"')
            else:
                suffix = "  # legacy" if was_legacy else ""
                lines.append(f'    {name} = "{value}"{suffix}')

    # Append new values not already in the enum.
    new_values = [v for v in api_values if v not in existing_values]
    if new_values:
        lines.append("")
        lines.append("    # New (auto-added by utils/check_hardware_flavors.py)")
        for value in new_values:
            lines.append(f'    {_make_enum_name(value)} = "{value}"')

    return lines


def _replace_enum_body(source: str, class_name: str, new_lines: list[str]) -> str:
    """Replace the member lines of an enum class in source code.

    Keeps everything up to and including the closing docstring, then replaces the
    enum member definitions up to the next top-level class/def/decorator.
    """
    # Step 1: find the enum class and its docstring.
    class_pattern = re.compile(rf"^class {re.escape(class_name)}\(str, Enum\):", re.MULTILINE)
    class_match = class_pattern.search(source)
    if not class_match:
        print(f"Error: could not find class {class_name}.")
        sys.exit(1)

    # Find the closing `"""` of the docstring (third occurrence of `"""` after class start).
    pos = class_match.start()
    doc_open = source.index('"""', pos)
    doc_close = source.index('"""', doc_open + 3) + 3
    header = source[:doc_close]

    # Step 2: find where the enum body ends (next top-level class/def/decorator).
    end_pattern = re.compile(r"^\n(?=class |def |@dataclass)", re.MULTILINE)
    end_match = end_pattern.search(source, doc_close)
    if not end_match:
        # Enum is at the end of file
        body_end = len(source)
    else:
        body_end = end_match.start()

    new_body = "\n".join(["", "", *new_lines, ""])
    return header + new_body + source[body_end:]


def _build_hardware_table() -> str:
    """Build a Markdown table by running ``hf jobs hardware --format json``."""
    result = subprocess.run(
        [sys.executable, "-m", "huggingface_hub.cli.hf", "jobs", "hardware", "--format", "json"],
        capture_output=True,
        text=True,
    )
    if result.returncode != 0:
        print(f"Error running `hf jobs hardware`:\n{result.stderr}")
        sys.exit(1)
    items: list[dict[str, str]] = json.loads(result.stdout)
    headers = list(items[0].keys())
    lines = [
        HARDWARE_TABLE_START,
        "| " + " | ".join(headers) + " |",
        "| " + " | ".join("---" for _ in headers) + " |",
    ]
    for item in items:
        row = [f"`{v}`" if k == "name" else (v or "-") for k, v in item.items()]
        lines.append("| " + " | ".join(row) + " |")
    lines.append(HARDWARE_TABLE_END)
    return "\n".join(lines)


def _replace_hardware_table(source: str, new_table: str) -> str:
    """Replace the hardware table block in the jobs guide."""
    start = source.find(HARDWARE_TABLE_START)
    end = source.find(HARDWARE_TABLE_END)
    if start == -1 or end == -1:
        print(f"Error: could not find hardware table markers in {JOBS_GUIDE_PATH}.")
        sys.exit(1)
    return source[:start] + new_table + source[end + len(HARDWARE_TABLE_END) :]


def check_hardware_flavors(update: bool) -> NoReturn:
    api = HfApi()

    # Fetch current hardware from the API.
    print("Fetching hardware flavors from the Hub API...")
    spaces_hardware = api.list_spaces_hardware()
    jobs_hardware = api.list_jobs_hardware()
    spaces_api_values = [hw.name for hw in spaces_hardware]
    jobs_api_values = [hw.name for hw in jobs_hardware]
    print(f"  Spaces: {len(spaces_api_values)} flavors")
    print(f"  Jobs:   {len(jobs_api_values)} flavors")

    # Read source files.
    space_source = SPACE_API_PATH.read_text()
    jobs_source = JOBS_API_PATH.read_text()
    jobs_guide_source = JOBS_GUIDE_PATH.read_text()

    # Parse existing enums.
    space_entries = _parse_enum_entries(space_source, "SpaceHardware")
    job_entries = _parse_enum_entries(jobs_source, "JobHardware")
    space_existing = {e.value for e in space_entries if e.kind == "member"}
    job_existing = {e.value for e in job_entries if e.kind == "member"}

    # Compute diffs.
    spaces_new = [v for v in spaces_api_values if v not in space_existing]
    jobs_new = [v for v in jobs_api_values if v not in job_existing]
    spaces_removed = [
        e.value
        for e in space_entries
        if e.kind == "member" and e.value not in set(spaces_api_values) and not e.is_legacy
    ]
    jobs_removed = [
        e.value for e in job_entries if e.kind == "member" and e.value not in set(jobs_api_values) and not e.is_legacy
    ]

    # Check if the hardware table in the docs needs updating.
    new_hardware_table = _build_hardware_table()
    new_guide_source = _replace_hardware_table(jobs_guide_source, new_hardware_table)
    guide_changed = new_guide_source != jobs_guide_source

    has_changes = bool(spaces_new or jobs_new or spaces_removed or jobs_removed or guide_changed)

    if spaces_new:
        print(f"\n  SpaceHardware — new:     {spaces_new}")
    if spaces_removed:
        print(f"  SpaceHardware — removed: {spaces_removed} (will be marked as legacy)")
    if jobs_new:
        print(f"  JobHardware — new:     {jobs_new}")
    if jobs_removed:
        print(f"  JobHardware — removed: {jobs_removed} (will be marked as legacy)")
    if guide_changed:
        print("\n  Jobs guide — hardware table is out of date")

    if not has_changes:
        print("\n✅ All good! Hardware enums and docs are up to date.")
        sys.exit(0)

    if not update:
        print(
            "\n❌ Hardware enums/docs are out of date.\n   Please run `python utils/check_hardware_flavors.py --update`."
        )
        sys.exit(1)

    # Update enum files.
    new_space_lines = _build_enum_lines(space_entries, spaces_api_values)
    new_job_lines = _build_enum_lines(job_entries, jobs_api_values)

    new_space_source = _replace_enum_body(space_source, "SpaceHardware", new_space_lines)
    new_jobs_source = _replace_enum_body(jobs_source, "JobHardware", new_job_lines)

    SPACE_API_PATH.write_text(new_space_source)
    JOBS_API_PATH.write_text(new_jobs_source)

    # Update docs.
    JOBS_GUIDE_PATH.write_text(new_guide_source)

    print("\n✅ Hardware enums and docs updated.\n   Please review the changes and commit them.")
    sys.exit(0)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Check/update hardware flavor enums against the Hub API.")
    parser.add_argument(
        "--update",
        action="store_true",
        help="Update the enum source files if changes are detected.",
    )
    args = parser.parse_args()
    check_hardware_flavors(update=args.update)



import base64
import time
from abc import ABC
from typing import Any

from huggingface_hub.hf_api import InferenceProviderMapping
from huggingface_hub.inference._common import (
    RequestParameters,
    _as_dict,
    _as_url,
)
from huggingface_hub.inference._providers._common import (
    BaseConversationalTask,
    BaseTextGenerationTask,
    TaskProviderHelper,
    filter_none,
)
from huggingface_hub.utils import get_session, hf_raise_for_status, logging


logger = logging.get_logger(__name__)

_PROVIDER = "together"
_BASE_URL = "https://api.together.xyz"

# Polling interval for async video generation (in seconds).
_VIDEO_POLLING_INTERVAL = 2.0

# Upper bound on status polls (initial response may already be terminal; each further poll is one attempt).
_VIDEO_MAX_POLL_ATTEMPTS = 150  # ~5 minutes at _VIDEO_POLLING_INTERVAL

# Job statuses that mean "keep polling". Together returns "queued" before transitioning to
# "in_progress", so we must treat both as pending.
_VIDEO_PENDING_STATUSES = {"queued", "in_progress"}


class TogetherTask(TaskProviderHelper, ABC):
    """Base class for Together API tasks."""

    def __init__(self, task: str):
        super().__init__(provider=_PROVIDER, base_url=_BASE_URL, task=task)

    def _prepare_route(self, mapped_model: str, api_key: str) -> str:
        match self.task:
            case "text-to-image" | "image-to-image":
                return "/v1/images/generations"
            case "text-to-speech":
                return "/v1/audio/speech"
            case "feature-extraction":
                return "/v1/embeddings"
            case "text-to-video" | "image-to-video":
                # Video creation lives under /v2 (see https://docs.together.ai/reference/create-videos).
                return "/v2/videos"
        raise ValueError(f"Unsupported task '{self.task}' for Together API.")


class TogetherTextGenerationTask(BaseTextGenerationTask):
    def __init__(self):
        super().__init__(provider=_PROVIDER, base_url=_BASE_URL)

    def get_response(self, response: bytes | dict, request_params: RequestParameters | None = None) -> Any:
        output = _as_dict(response)["choices"][0]
        return {
            "generated_text": output["text"],
            "details": {
                "finish_reason": output.get("finish_reason"),
                "seed": output.get("seed"),
            },
        }


class TogetherConversationalTask(BaseConversationalTask):
    def __init__(self):
        super().__init__(provider=_PROVIDER, base_url=_BASE_URL)

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        payload = super()._prepare_payload_as_dict(inputs, parameters, provider_mapping_info)
        if payload is None:
            return None
        # Together accepts response_format `{type: "json_schema", schema: <schema>}` (flattened),
        # so unwrap the OpenAI-style `{type: "json_schema", json_schema: {schema}}` envelope.
        response_format = payload.get("response_format")
        if (
            isinstance(response_format, dict)
            and response_format.get("type") == "json_schema"
            and isinstance(response_format.get("json_schema"), dict)
            and "schema" in response_format["json_schema"]
        ):
            payload["response_format"] = {
                "type": "json_schema",
                "schema": response_format["json_schema"]["schema"],
            }
        return payload


class TogetherTextToImageTask(TogetherTask):
    def __init__(self):
        super().__init__("text-to-image")

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        mapped_model = provider_mapping_info.provider_id
        parameters = filter_none(parameters)
        if "num_inference_steps" in parameters:
            parameters["steps"] = parameters.pop("num_inference_steps")

        return {"prompt": inputs, "response_format": "base64", **parameters, "model": mapped_model}

    def get_response(self, response: bytes | dict, request_params: RequestParameters | None = None) -> Any:
        response_dict = _as_dict(response)
        return base64.b64decode(response_dict["data"][0]["b64_json"])


class TogetherImageToImageTask(TogetherTask):
    def __init__(self):
        super().__init__("image-to-image")

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        mapped_model = provider_mapping_info.provider_id
        image_url = _as_url(inputs, default_mime_type="image/jpeg")

        # Filter `None` values first: the client always passes `"prompt": None` when the user
        # omits the argument, so popping before filtering would yield `None` instead of the
        # `""` default and send `"prompt": null` to Together (rejected by Flux Kontext).
        parameters = filter_none(parameters)
        prompt = parameters.pop("prompt", "")
        if "num_inference_steps" in parameters:
            parameters["steps"] = parameters.pop("num_inference_steps")

        # Together exposes two mutually-exclusive image inputs (see
        # https://docs.together.ai/docs/image-to-image): FLUX.1 Kontext only accepts
        # `image_url`; FLUX.2 [dev] and Google models (Gemini 3 Pro Image, Flash Image
        # 2.5) only accept `reference_images`. FLUX.2 [pro]/[flex] accept either but
        # `reference_images` is the documented default. Use `image_url` only for
        # FLUX.1 Kontext models and `reference_images` for everything else.
        lowered = mapped_model.lower()
        use_image_url = "kontext" in lowered and "flux.1" in lowered
        image_field: dict[str, Any] = {"image_url": image_url} if use_image_url else {"reference_images": [image_url]}
        return {
            "prompt": prompt,
            **image_field,
            "response_format": "base64",
            **parameters,
            "model": mapped_model,
        }

    def get_response(self, response: bytes | dict, request_params: RequestParameters | None = None) -> Any:
        response_dict = _as_dict(response)
        return base64.b64decode(response_dict["data"][0]["b64_json"])


class TogetherFeatureExtractionTask(TogetherTask):
    def __init__(self):
        super().__init__("feature-extraction")

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        return {
            "input": inputs,
            "model": provider_mapping_info.provider_id,
            **filter_none(parameters),
        }

    def get_response(self, response: bytes | dict, request_params: RequestParameters | None = None) -> Any:
        return [item["embedding"] for item in _as_dict(response)["data"]]


class TogetherTextToSpeechTask(TogetherTask):
    def __init__(self):
        super().__init__("text-to-speech")

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        # `voice` is required by the Together API and is model-specific
        # (see https://docs.together.ai/docs/text-to-speech#supported-voices),
        # so we don't set a default and let the API surface a clear error if missing.
        return {
            "input": inputs,
            "model": provider_mapping_info.provider_id,
            **filter_none(parameters),
        }

    def get_response(self, response: bytes | dict, request_params: RequestParameters | None = None) -> Any:
        if isinstance(response, bytes):
            return response
        raise ValueError(f"Expected raw audio bytes for text-to-speech, got {type(response).__name__}.")


def _normalize_video_parameters(parameters: dict) -> dict:
    """Map HF inference-client conventions onto Together's video API parameter names."""
    parameters = filter_none(parameters)
    if "num_inference_steps" in parameters:
        parameters["steps"] = parameters.pop("num_inference_steps")
    if "target_size" in parameters:
        target_size = parameters.pop("target_size")
        if "width" in target_size:
            parameters["width"] = target_size["width"]
        if "height" in target_size:
            parameters["height"] = target_size["height"]
    return parameters


class TogetherVideoTask(TogetherTask, ABC):
    """Base class for Together's asynchronous video generation tasks."""

    def get_response(self, response: bytes | dict, request_params: RequestParameters | None = None) -> Any:
        if request_params is None:
            raise ValueError("A `RequestParameters` object is required to poll Together video jobs.")

        job = _as_dict(response)
        job_id = job.get("id")
        if not job_id:
            raise ValueError("No job ID found in Together video generation response.")

        # Status polling lives at the same /v2/videos URL with the job ID appended.
        status_url = f"{request_params.url}/{job_id}"

        logger.info("Generating video, polling for completion...")
        # Together usually returns `status: "queued"` on the initial POST, but the field is
        # optional per the spec — treat a missing status as "still pending" and poll, rather
        # than falling through to the "unexpected status" error below.
        status = job.get("status")
        for _ in range(_VIDEO_MAX_POLL_ATTEMPTS):
            if status is not None and status not in _VIDEO_PENDING_STATUSES:
                break
            time.sleep(_VIDEO_POLLING_INTERVAL)
            status_response = get_session().get(status_url, headers=request_params.headers)
            hf_raise_for_status(status_response)
            job = status_response.json()
            status = job.get("status")
            if status is not None and status not in _VIDEO_PENDING_STATUSES:
                break
        else:
            raise ValueError(
                "Timed out while waiting for Together video generation "
                f"— aborting after {_VIDEO_MAX_POLL_ATTEMPTS} status polls"
            )

        if status == "failed":
            error = job.get("error") or {}
            raise RuntimeError(f"Together video generation failed: {error.get('message') or 'Unknown error'}")
        if status != "completed":
            raise RuntimeError(f"Unexpected Together video job status: {status!r}")

        video_url = (job.get("outputs") or {}).get("video_url")
        if not video_url:
            raise ValueError("No video URL found in completed Together video job.")

        video_response = get_session().get(video_url)
        hf_raise_for_status(video_response)
        return video_response.content


class TogetherTextToVideoTask(TogetherVideoTask):
    def __init__(self):
        super().__init__("text-to-video")

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        return {
            "prompt": inputs,
            "model": provider_mapping_info.provider_id,
            **_normalize_video_parameters(parameters),
        }


class TogetherImageToVideoTask(TogetherVideoTask):
    def __init__(self):
        super().__init__("image-to-video")

    def _prepare_payload_as_dict(
        self, inputs: Any, parameters: dict, provider_mapping_info: InferenceProviderMapping
    ) -> dict | None:
        # Together expects each keyframe as `{input_image, frame: "first" | "last"}`
        # for i2v models. See https://docs.together.ai/docs/inference/videos/reference-and-keyframes.
        # Note: `input_image` accepts a data URL or an HTTP(S) URL but the field is capped
        # at ~60KB — users with larger inputs should host the image and pass `frame_images`
        # directly via `extra_body`.
        return {
            "model": provider_mapping_info.provider_id,
            "frame_images": [{"input_image": _as_url(inputs, default_mime_type="image/png"), "frame": "first"}],
            **_normalize_video_parameters(parameters),
        }



from __future__ import annotations

import asyncio
from typing import AsyncGenerator, Iterable, Optional, Union

from huggingface_hub import ChatCompletionInputMessage, ChatCompletionStreamOutput, MCPClient

from .._providers import PROVIDER_OR_POLICY_T
from .constants import DEFAULT_SYSTEM_PROMPT, EXIT_LOOP_TOOLS, MAX_NUM_TURNS
from .types import ServerConfig


class Agent(MCPClient):
    """
    Implementation of a Simple Agent, which is a simple while loop built right on top of an [`MCPClient`].

    > [!WARNING]
    > This class is experimental and might be subject to breaking changes in the future without prior notice.

    Args:
        model (`str`, *optional*):
            The model to run inference with. Can be a model id hosted on the Hugging Face Hub, e.g. `meta-llama/Meta-Llama-3-8B-Instruct`
            or a URL to a deployed Inference Endpoint or other local or remote endpoint.
        servers (`Iterable[dict]`):
            MCP servers to connect to. Each server is a dictionary containing a `type` key and a `config` key. The `type` key can be `"stdio"` or `"sse"`, and the `config` key is a dictionary of arguments for the server.
        provider (`str`, *optional*):
            Name of the provider to use for inference. Defaults to "auto" i.e. the first of the providers available for the model, sorted by the user's order in https://hf.co/settings/inference-providers.
            If model is a URL or `base_url` is passed, then `provider` is not used.
        base_url (`str`, *optional*):
            The base URL to run inference. Defaults to None.
        api_key (`str`, *optional*):
            Token to use for authentication. Will default to the locally Hugging Face saved token if not provided. You can also use your own provider API key to interact directly with the provider's service.
        prompt (`str`, *optional*):
            The system prompt to use for the agent. Defaults to the default system prompt in `constants.py`.
    """

    def __init__(
        self,
        *,
        model: Optional[str] = None,
        servers: Iterable[ServerConfig],
        provider: Optional[PROVIDER_OR_POLICY_T] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        prompt: Optional[str] = None,
    ):
        super().__init__(model=model, provider=provider, base_url=base_url, api_key=api_key)
        self._servers_cfg = list(servers)
        self.messages: list[Union[dict, ChatCompletionInputMessage]] = [
            {"role": "system", "content": prompt or DEFAULT_SYSTEM_PROMPT}
        ]

    async def load_tools(self) -> None:
        for cfg in self._servers_cfg:
            await self.add_mcp_server(**cfg)

    async def run(
        self,
        user_input: str,
        *,
        abort_event: Optional[asyncio.Event] = None,
    ) -> AsyncGenerator[Union[ChatCompletionStreamOutput, ChatCompletionInputMessage], None]:
        """
        Run the agent with the given user input.

        Args:
            user_input (`str`):
                The user input to run the agent with.
            abort_event (`asyncio.Event`, *optional*):
                An event that can be used to abort the agent. If the event is set, the agent will stop running.
        """
        self.messages.append({"role": "user", "content": user_input})

        num_turns: int = 0
        next_turn_should_call_tools = True

        while True:
            if abort_event and abort_event.is_set():
                return

            async for item in self.process_single_turn_with_tools(
                self.messages,
                exit_loop_tools=EXIT_LOOP_TOOLS,
                exit_if_first_chunk_no_tool=(num_turns > 0 and next_turn_should_call_tools),
            ):
                yield item

            num_turns += 1
            last = self.messages[-1]

            if last.get("role") == "tool" and last.get("name") in {t.function.name for t in EXIT_LOOP_TOOLS}:
                return

            if last.get("role") != "tool" and num_turns > MAX_NUM_TURNS:
                return

            if last.get("role") != "tool" and next_turn_should_call_tools:
                return

            next_turn_should_call_tools = last.get("role") != "tool"
